"""Contains functions for getting market level data."""
from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.urls import *
from robin_stocks.robinhood.stocks import *

def get_top_movers_sp500(direction, info=None):
    """Returns a list of the top S&P500 movers up or down for the day.

    :param direction: The direction of movement either 'up' or 'down'
    :type direction: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each mover. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * instrument_url
                      * symbol
                      * updated_at
                      * price_movement
                      * description

    """
    try:
        direction = direction.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    if (direction != 'up' and direction != 'down'):
        print('Error: direction must be "up" or "down"', file=get_output())
        return([None])

    url = movers_sp500_url()
    payload = {'direction': direction}
    data = request_get(url, 'pagination', payload)

    return(filter_data(data, info))

def get_top_100(info=None):
    """Returns a list of the Top 100 stocks on Robin Hood.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each mover. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * ask_price
                      * ask_size
                      * bid_price
                      * bid_size
                      * last_trade_price
                      * last_extended_hours_trade_price
                      * previous_close
                      * adjusted_previous_close
                      * previous_close_date
                      * symbol
                      * trading_halted
                      * has_traded
                      * last_trade_price_source
                      * updated_at
                      * instrument

    """
    url = get_100_most_popular_url()
    data = request_get(url, 'regular')
    data = filter_data(data, 'instruments')

    symbols = [get_symbol_by_url(x) for x in data]
    data = get_quotes(symbols)

    return(filter_data(data, info))

def get_top_movers(info=None):
    """Returns a list of the Top 20 movers on Robin Hood.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each mover. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * ask_price
                      * ask_size
                      * bid_price
                      * bid_size
                      * last_trade_price
                      * last_extended_hours_trade_price
                      * previous_close
                      * adjusted_previous_close
                      * previous_close_date
                      * symbol
                      * trading_halted
                      * has_traded
                      * last_trade_price_source
                      * updated_at
                      * instrument

    """
    url = movers_top_url()
    data = request_get(url, 'regular')
    data = filter_data(data, 'instruments')

    symbols = [get_symbol_by_url(x) for x in data]
    data = get_quotes(symbols)

    return(filter_data(data, info))

def get_all_stocks_from_market_tag(tag, info=None):
    """Returns all the stock quote information that matches a tag category.

    :param tag: The category to filter for. Examples include 'biopharmaceutical', 'upcoming-earnings', 'most-popular-under-25', and 'technology'.
    :type tag: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each mover. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * ask_price
                      * ask_size
                      * bid_price
                      * bid_size
                      * last_trade_price
                      * last_extended_hours_trade_price
                      * previous_close
                      * adjusted_previous_close
                      * previous_close_date
                      * symbol
                      * trading_halted
                      * has_traded
                      * last_trade_price_source
                      * updated_at
                      * instrument

    """
    url = market_category_url(tag)
    data = request_get(url, 'regular')
    data = filter_data(data, 'instruments')

    if not data:
        print('ERROR: "{}" is not a valid tag'.format(tag), file=get_output())
        return [None]

    symbols = [get_symbol_by_url(x) for x in data]
    data = get_quotes(symbols)

    return(filter_data(data, info))

def get_markets(info=None):
    """Returns a list of available markets.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each market. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * url
                      * todays_hours
                      * mic
                      * operating_mic
                      * acronym
                      * name
                      * city
                      * country
                      * timezone
                      * website

    """
    url = markets_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))

def get_market_today_hours(market, info=None):
    """Returns the opening and closing hours of a specific market for today. Also will tell you if market is
    market is open on that date.

    :param market: The 'mic' value for the market. Can be found using get_markets().
    :type market: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the specific market. If info parameter is provided, \
    the string value for the corresponding key will be provided.
    :Dictionary Keys: * date
                      * is_open
                      * opens_at
                      * closes_at
                      * extended_opens_at
                      * extended_closes_at
                      * previous_open_hours
                      * next_open_hours

    """
    markets = get_markets()
    result = next((x for x in markets if x['mic'] == market), None)
    if not result:
        raise Exception('Not a valid market name. Check get_markets() for a list of market information.')

    url = result['todays_hours']
    data = request_get(url, 'regular')
    return(filter_data(data, info))

def get_market_next_open_hours(market, info=None):
    """Returns the opening and closing hours for the next open trading day after today. Also will tell you if market is
    market is open on that date.

    :param market: The 'mic' value for the market. Can be found using get_markets().
    :type market: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the specific market. If info parameter is provided, \
    the string value for the corresponding key will be provided.
    :Dictionary Keys: * date
                      * is_open
                      * opens_at
                      * closes_at
                      * extended_opens_at
                      * extended_closes_at
                      * previous_open_hours
                      * next_open_hours

    """
    url = get_market_today_hours(market, info='next_open_hours')
    data = request_get(url, 'regular')
    return(filter_data(data, info))

def get_market_next_open_hours_after_date(market, date, info=None):
    """Returns the opening and closing hours for the next open trading day after a date that is specified. Also will tell you if market is
    market is open on that date.

    :param market: The 'mic' value for the market. Can be found using get_markets().
    :type market: str
    :param date: The date you want to find the next available trading day after. format is YYYY-MM-DD.
    :type date: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the specific market. If info parameter is provided, \
    the string value for the corresponding key will be provided.
    :Dictionary Keys: * date
                      * is_open
                      * opens_at
                      * closes_at
                      * extended_opens_at
                      * extended_closes_at
                      * previous_open_hours
                      * next_open_hours

    """
    url = get_market_hours(market, date, info='next_open_hours')
    data = request_get(url, 'regular')
    return(filter_data(data, info))

def get_market_hours(market, date, info=None):
    """Returns the opening and closing hours of a specific market on a specific date. Also will tell you if market is
    market is open on that date. Can be used with past or future dates.

    :param market: The 'mic' value for the market. Can be found using get_markets().
    :type market: str
    :param date: The date you want to get information for. format is YYYY-MM-DD.
    :type date: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the specific market. If info parameter is provided, \
    the string value for the corresponding key will be provided.
    :Dictionary Keys: * date
                      * is_open
                      * opens_at
                      * closes_at
                      * extended_opens_at
                      * extended_closes_at
                      * previous_open_hours
                      * next_open_hours

    """
    url = market_hours_url(market, date)
    data = request_get(url, 'regular')
    return(filter_data(data, info))


def get_currency_pairs(info=None):
    """Returns currency pairs

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each currency pair. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * asset_currency
                      * display_only
                      * id
                      * max_order_size
                      * min_order_size
                      * min_order_price_increment
                      * min_order_quantity_increment
                      * name
                      * quote_currency
                      * symbol
                      * tradability

    """
    url = currency_url()
    data = request_get(url, 'results')
    return(filter_data(data, info))



# Used by git Actions
import os
import datetime
import robin_stocks.robinhood as r
import pyotp
import pytest
from dateutil.relativedelta import relativedelta

def third_friday(year, month, day):
    """Return datetime.date for monthly option expiration given year and
    month
    """
    # The 15th is the lowest third day in the month
    third = datetime.date(year, month, 15)
    # What day of the week is the 15th?
    w = third.weekday()
    # Friday is weekday 4
    if w != 4:
        # Replace just the day (of month)
        third = third.replace(day=(15 + (4 - w) % 7))

    if day > third.day:
        month += 1
        third = datetime.date(year, month, 15)
        w = third.weekday()
        if w != 4:
            third = third.replace(day=(15 + (4 - w) % 7))

    return third


def round_up_price(ticker, multiplier):
    price = float(r.get_latest_price(ticker)[0])
    num = price + (multiplier - 1)
    return num - (num % multiplier)

class TestStocks:

    # Set up variables for class
    single_stock = 'AAPL'
    event_stock = 'USO1'
    fake_stock = 'thisisfake'
    instrument = 'https://api.robinhood.com/instruments/450dfc6d-5510-4d40-abfb-f633b7d9be3e/'
    fake_instrument = 'https://api.robinhood.com/instruments/aaaaaaaa-0000-0000-0000-aaaaaaaaaaaa/'
    id = '450dfc6d-5510-4d40-abfb-f633b7d9be3e'
    list_stocks = ['tsla', 'f', 'plug', 'meTA', 'SPY', 'botz', 'jnug']
    fake_stocks = ['87627273', 'ffffffffff']

    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)

    @classmethod
    def teardown_class(cls):
        r.logout()

    def test_name_apple(self):
        name = r.get_name_by_symbol('aapl')
        assert name == "Apple"

    def test_quotes(self):
        quote = r.get_quotes(self.single_stock, info=None)
        assert (len(quote) == 1)
        quote = quote[0]
        assert (quote['symbol'] == self.single_stock)
        assert ('ask_price' in quote)
        assert ('ask_size' in quote)
        assert ('bid_price' in quote)
        assert ('bid_size' in quote)
        assert ('last_trade_price' in quote)
        assert ('last_extended_hours_trade_price' in quote)
        assert ('previous_close' in quote)
        assert ('adjusted_previous_close' in quote)
        assert ('previous_close_date' in quote)
        assert ('symbol' in quote)
        assert ('trading_halted' in quote)
        assert ('has_traded' in quote)
        assert ('last_trade_price_source' in quote)
        assert ('updated_at' in quote)
        assert ('instrument' in quote)
        #
        more_quotes = r.get_quotes(self.list_stocks, info=None)
        assert (len(more_quotes) ==  len(self.list_stocks))
        #
        fake_quotes = r.get_quotes(self.fake_stocks, info=None)
        assert (len(fake_quotes) == 1)
        assert (fake_quotes[0] == None)

    def test_fundamentals(self):
        quote = r.get_fundamentals(self.single_stock, info=None)
        assert (len(quote) == 1)
        assert (quote[0]['symbol'] == self.single_stock)
        quote = quote[0]
        assert ('open' in quote)
        assert ('high' in quote)
        assert ('low' in quote)
        assert ('volume' in quote)
        assert ('average_volume_2_weeks' in quote)
        assert ('average_volume' in quote)
        assert ('high_52_weeks' in quote)
        assert ('dividend_yield' in quote)
        assert ('float' in quote)
        assert ('low_52_weeks' in quote)
        assert ('market_cap' in quote)
        assert ('pb_ratio' in quote)
        assert ('pe_ratio' in quote)
        assert ('shares_outstanding' in quote)
        assert ('description' in quote)
        assert ('instrument' in quote)
        assert ('ceo' in quote)
        assert ('headquarters_city' in quote)
        assert ('headquarters_state' in quote)
        assert ('sector' in quote)
        assert ('industry' in quote)
        assert ('num_employees' in quote)
        assert ('year_founded' in quote)
        assert ('symbol' in quote)
        #
        more_quotes = r.get_fundamentals(self.list_stocks, info=None)
        assert (len(more_quotes) == len(self.list_stocks))
        #
        fake_quotes = r.get_fundamentals(self.fake_stocks, info=None)
        assert (len(fake_quotes) == 1)
        assert (fake_quotes[0] == None)

    def test_instruments(self):
        quote = r.get_instruments_by_symbols(self.single_stock)
        assert (len(quote) == 1)
        assert (quote[0]['symbol'] == self.single_stock)
        quote = quote[0]
        assert ('id' in quote)
        assert ('url' in quote)
        assert ('quote' in quote)
        assert ('fundamentals' in quote)
        assert ('splits' in quote)
        assert ('state' in quote)
        assert ('market' in quote)
        assert ('simple_name' in quote)
        assert ('name' in quote)
        assert ('tradeable' in quote)
        assert ('tradability' in quote)
        assert ('symbol' in quote)
        assert ('bloomberg_unique' in quote)
        assert ('margin_initial_ratio' in quote)
        assert ('maintenance_ratio' in quote)
        assert ('country' in quote)
        assert ('day_trade_ratio' in quote)
        assert ('list_date' in quote)
        assert ('min_tick_size' in quote)
        assert ('type' in quote)
        assert ('tradable_chain_id' in quote)
        assert ('rhs_tradability' in quote)
        assert ('fractional_tradability' in quote)
        assert ('default_collar_fraction' in quote)
        #
        more_quotes = r.get_fundamentals(self.list_stocks, info=None)
        assert (len(more_quotes) == len(self.list_stocks))
        #
        fake_quotes = r.get_fundamentals(self.fake_stocks, info=None)
        assert (len(fake_quotes) == 1)
        assert (fake_quotes[0] == None)

    def test_instrument_id(self):
        quote = r.get_instrument_by_url(self.instrument)
        assert (quote['symbol'] == self.single_stock)
        assert ('id' in quote)
        assert ('url' in quote)
        assert ('quote' in quote)
        assert ('fundamentals' in quote)
        assert ('splits' in quote)
        assert ('state' in quote)
        assert ('market' in quote)
        assert ('simple_name' in quote)
        assert ('name' in quote)
        assert ('tradeable' in quote)
        assert ('tradability' in quote)
        assert ('symbol' in quote)
        assert ('bloomberg_unique' in quote)
        assert ('margin_initial_ratio' in quote)
        assert ('maintenance_ratio' in quote)
        assert ('country' in quote)
        assert ('day_trade_ratio' in quote)
        assert ('list_date' in quote)
        assert ('min_tick_size' in quote)
        assert ('type' in quote)
        assert ('tradable_chain_id' in quote)
        assert ('rhs_tradability' in quote)
        assert ('fractional_tradability' in quote)
        assert ('default_collar_fraction' in quote)

    def test_latest_price(self):
        price = r.get_latest_price(self.single_stock)
        assert (len(price) == 1)
        more_prices = r.get_latest_price(self.list_stocks)
        assert (len(more_prices) == len(self.list_stocks))
        fake_prices = r.get_latest_price(self.fake_stocks)
        assert (len(fake_prices) == 1)
        assert (fake_prices[0] == None)

    def test_name_by_symbol(self):
        name = r.get_name_by_symbol(self.single_stock)
        assert (name == 'Apple')
        fake_name = r.get_name_by_symbol(self.fake_stock)
        assert (fake_name == '')

    def test_name_by_url(self):
        name = r.get_name_by_url(self.instrument)
        assert (name == 'Apple')
        fake_name = r.get_name_by_url(self.fake_instrument)
        assert (fake_name == '')

    def test_symbol_by_url(self):
        symbol = r.get_symbol_by_url(self.instrument)
        assert (symbol == self.single_stock)
        fake_symbol = r.get_symbol_by_url(self.fake_instrument)
        assert (fake_symbol == '')

    def test_get_ratings(self):
        ratings = r.get_ratings(self.single_stock)
        assert ('summary' in ratings)
        assert ('ratings' in ratings)
        assert ('instrument_id' in ratings)
        assert ('ratings_published_at' in ratings)
        fake_ratings = r.get_ratings(self.fake_stock)
        assert (fake_ratings == '')

    def test_events(self):
        event = r.get_events(self.single_stock)
        assert (len(event) == 0)
        event = r.get_events(self.event_stock)
        assert (len(event) != 0)
        event = event[0]
        assert ('account' in event)
        assert ('cash_component' in event)
        assert ('chain_id' in event)
        assert ('created_at' in event)
        assert ('direction' in event)
        assert ('equity_components' in event)
        assert ('event_date' in event)
        assert ('id' in event)
        assert ('option' in event)
        assert ('position' in event)
        assert ('quantity' in event)
        assert ('state' in event)
        assert ('total_cash_amount' in event)
        assert ('type' in event)
        assert ('underlying_price' in event)
        assert ('updated_at' in event)

    def test_earning(self):
        earnings = r.get_earnings(self.single_stock)[0]
        assert (earnings['symbol'] == self.single_stock)
        assert ('symbol' in earnings)
        assert ('instrument' in earnings)
        assert ('year' in earnings)
        assert ('quarter' in earnings)
        assert ('eps' in earnings)
        assert ('report' in earnings)
        assert ('call' in earnings)
        fake_earnings = r.get_earnings(self.fake_stock)
        assert (len(fake_earnings) == 0)

    def test_news(self):
        news = r.get_news(self.single_stock)[0]
        assert ('author' in news)
        assert ('num_clicks' in news)
        assert ('preview_image_url' in news)
        assert ('published_at' in news)
        assert ('relay_url' in news)
        assert ('source' in news)
        assert ('summary' in news)
        assert ('title' in news)
        assert ('updated_at' in news)
        assert ('url' in news)
        assert ('uuid' in news)
        assert ('related_instruments' in news)
        assert ('preview_text' in news)
        assert ('currency_id' in news)
        fake_news = r.get_news(self.fake_stock)
        assert (len(fake_news) == 0)

    def test_split(self):
        split = r.get_splits(self.single_stock)[0]
        assert (split['instrument'] == self.instrument)
        fake_split = r.get_splits(self.fake_stock)
        assert (len(fake_split) == 0)

    def test_stock_historicals(self):
        historicals = r.get_stock_historicals(self.single_stock, interval='hour', span='day', bounds='regular', info=None)
        assert (len(historicals) <= 6) # 6 regular hours in a day
        historicals = r.get_stock_historicals(self.single_stock, interval='hour', span='day', bounds='trading', info=None)
        assert (len(historicals) <= 9) # 9 trading hours total in a day
        historicals = r.get_stock_historicals(self.single_stock, interval='hour', span='day', bounds='extended', info=None)
        assert (len(historicals) <= 16) # 16 extended hours total in a day

class TestCrypto:

    stock = 'AAPL'
    bitcoin = 'BTC'
    bitcoin_currency = 'BTC-USD'
    bitcoin_symbol = 'BTCUSD'
    fake = 'thisisafake'
    account = os.environ['crypto_account']

    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)

    @classmethod
    def teardown_class(cls):
        r.logout()

    def test_crypto_positions(self):
        positions = r.get_crypto_positions(info=None)
        first = positions[0]
        assert (first['account_id'] == self.account)
        assert ('account_id' in first)
        assert ('cost_bases' in first)
        assert ('created_at' in first)
        assert ('currency' in first)
        assert ('id' in first)
        assert ('quantity' in first)
        assert ('quantity_available' in first)
        assert ('quantity_held_for_buy' in first)
        assert ('quantity_held_for_sell' in first)
        assert ('updated_at' in first)

    def test_get_crypto_currency_pairs(self):
        pairs = r.get_crypto_currency_pairs(info=None)
        btc = [x for x in pairs if x['symbol'] == self.bitcoin_currency][0]
        assert ('asset_currency' in btc)
        assert ('display_only' in btc)
        assert ('id' in btc)
        assert ('max_order_size' in btc)
        assert ('min_order_size' in btc)
        assert ('min_order_price_increment' in btc)
        assert ('min_order_quantity_increment' in btc)
        assert ('name' in btc)
        assert ('quote_currency' in btc)
        assert ('symbol' in btc)
        assert ('tradability' in btc)
        fake = [x for x in pairs if x['symbol'] == self.fake]
        assert (len(fake) == 0)

    def test_crypto_info(self):
        crypto = r.get_crypto_info(self.bitcoin, info=None)
        assert ('asset_currency' in crypto)
        assert ('display_only' in crypto)
        assert ('id' in crypto)
        assert ('max_order_size' in crypto)
        assert ('min_order_size' in crypto)
        assert ('min_order_price_increment' in crypto)
        assert ('min_order_quantity_increment' in crypto)
        assert ('name' in crypto)
        assert ('quote_currency' in crypto)
        assert ('symbol' in crypto)
        assert ('tradability' in crypto)
        crypto = r.get_crypto_info(self.stock, info=None)
        assert (crypto == None)

    def test_crypto_quote(self):
        crypto = r.get_crypto_quote(self.bitcoin, info=None)
        assert ('ask_price' in crypto)
        assert ('bid_price' in crypto)
        assert ('mark_price' in crypto)
        assert ('high_price' in crypto)
        assert ('low_price' in crypto)
        assert ('open_price' in crypto)
        assert ('symbol' in crypto)
        assert ('id' in crypto)
        assert ('volume' in crypto)
        crypto = r.get_crypto_quote(self.stock, info=None)
        assert (crypto == None)
        crypto = r.get_crypto_quote(self.fake, info=None)
        assert (crypto == None)

    def test_crypto_historicals(self):
        crypto = r.get_crypto_historicals(self.bitcoin, 'day', 'week', '24_7', info=None)
        assert (len(crypto) == 7)
        first_point = crypto[0]
        # check keys
        assert ('begins_at' in first_point)
        assert ('open_price' in first_point)
        assert ('close_price' in first_point)
        assert ('high_price' in first_point)
        assert ('low_price' in first_point)
        assert ('volume' in first_point)
        assert ('session' in first_point)
        assert ('interpolated' in first_point)
        #
        crypto = r.get_crypto_historicals(self.bitcoin, 'hour', 'day', 'regular', info=None)
        assert (len(crypto) <= 6) # 6 regular hours in a day
        crypto = r.get_crypto_historicals(self.bitcoin, 'hour', 'day', 'trading', info=None)
        assert (len(crypto) <= 9) # 9 trading hours in a day
        crypto = r.get_crypto_historicals(self.bitcoin, 'hour', 'day', 'extended', info=None)
        assert (len(crypto) <= 16) # 16 extended hours in a day
        crypto = r.get_crypto_historicals(self.bitcoin, 'hour', 'day', '24_7', info=None)
        assert (len(crypto) <= 24) # 24 24_7 hours in a day

class TestOptions:

    # have to login to use round_up_price
    totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
    login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)
    #
    now = datetime.datetime.now() + relativedelta(months=1)
    expiration_date = third_friday(now.year, now.month, now.day).strftime("%Y-%m-%d")
    symbol = 'AAPL'
    strike = round_up_price(symbol, 10)

    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)

    @classmethod
    def teardown_class(cls):
        r.logout()

    def test_find_tradable_options(self):
        info = r.find_options_by_expiration(self.symbol, self.expiration_date)
        first = info[0]
        assert (first['expiration_date'] == self.expiration_date)
        assert (len(info) > 50)
        info = r.find_options_by_expiration(self.symbol, self.expiration_date, info='strike_price')
        first = info[0]
        assert (type(first) == str)
        assert (len(info) > 50)
        info = r.find_options_by_expiration(self.symbol, self.expiration_date, info='expiration_date')
        assert (len(set(info)) == 1)

    def test_find_options_by_strike(self):
        info = r.find_options_by_strike(self.symbol, self.strike)
        assert (len(info) >= 24)
        info = r.find_options_by_strike(self.symbol, self.strike,'call')
        assert (info[0]['type'] == 'call')
        info = r.find_options_by_strike(self.symbol, self.strike, info='expiration_date')
        assert (len(set(info)) > 1)
        info = r.find_options_by_strike(self.symbol, self.strike, info='strike_price')
        assert (len(set(info)) == 1)

    def test_find_options_by_expiration_and_strike(self):
        info = r.find_options_by_expiration_and_strike(self.symbol, self.expiration_date, self.strike)
        assert (len(info) == 2)
        assert (info[0]['expiration_date'] == self.expiration_date)
        assert (float(info[0]['strike_price']) == self.strike)
        info = r.find_options_by_expiration_and_strike(self.symbol, self.expiration_date, self.strike, 'call')
        assert (len(info) == 1)
        assert (info[0]['type'] == 'call')

class TestMarkets:

    today = datetime.datetime.today().strftime('%Y-%m-%d')
    american_time = datetime.datetime.today().strftime('%m-%d-%Y')
    nyse = 'XNYS'
    amex = 'XASE'
    nasdaq = 'XNAS'
    fake = 'blah'

    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)

    @classmethod
    def teardown_class(cls):
        r.logout()

    def test_top_movers(self):
        movers = r.get_top_movers()
        assert (movers)
        first = movers[0]
        assert ('ask_price' in first)
        assert ('ask_size' in first)
        assert ('bid_price' in first)
        assert ('bid_size' in first)
        assert ('last_trade_price' in first)
        assert ('last_extended_hours_trade_price' in first)
        assert ('previous_close' in first)
        assert ('adjusted_previous_close' in first)
        assert ('previous_close_date' in first)
        assert ('symbol' in first)
        assert ('trading_halted' in first)
        assert ('has_traded' in first)
        assert ('last_trade_price_source' in first)
        assert ('updated_at' in first)
        assert ('instrument' in first)

    def test_top_100(self):
        movers = r.get_top_100()
        assert (movers)
        first = movers[0]
        assert ('ask_price' in first)
        assert ('ask_size' in first)
        assert ('bid_price' in first)
        assert ('bid_size' in first)
        assert ('last_trade_price' in first)
        assert ('last_extended_hours_trade_price' in first)
        assert ('previous_close' in first)
        assert ('adjusted_previous_close' in first)
        assert ('previous_close_date' in first)
        assert ('symbol' in first)
        assert ('trading_halted' in first)
        assert ('has_traded' in first)
        assert ('last_trade_price_source' in first)
        assert ('updated_at' in first)
        assert ('instrument' in first)

    def test_top_movers_sp500(self):
        # going up
        movers = r.get_top_movers_sp500('up')
        assert (movers)
        first = movers[0]
        assert ('instrument_url' in first)
        assert ('symbol' in first)
        assert ('updated_at' in first)
        assert ('price_movement' in first)
        assert ('description' in first)
        assert ('market_hours_last_movement_pct' in first['price_movement'])
        assert ('market_hours_last_price' in first['price_movement'])
        assert (float(first['price_movement']['market_hours_last_movement_pct']) > 0)
        # going down
        movers = r.get_top_movers_sp500('down')
        assert (movers)
        first = movers[0]
        assert (float(first['price_movement']['market_hours_last_movement_pct']) < 0)

    def test_get_all_stocks_from_market_tag(self):
        movers = r.get_all_stocks_from_market_tag('technology')
        assert (movers)
        first = movers[0]
        assert ('ask_price' in first)
        assert ('ask_size' in first)
        assert ('bid_price' in first)
        assert ('bid_size' in first)
        assert ('last_trade_price' in first)
        assert ('last_extended_hours_trade_price' in first)
        assert ('previous_close' in first)
        assert ('adjusted_previous_close' in first)
        assert ('previous_close_date' in first)
        assert ('symbol' in first)
        assert ('trading_halted' in first)
        assert ('has_traded' in first)
        assert ('last_trade_price_source' in first)
        assert ('updated_at' in first)
        assert ('instrument' in first)

    def test_get_markets(self):
        markets = r.get_markets()
        assert (markets)
        first = markets[0]
        assert ('url' in first)
        assert ('todays_hours' in first)
        assert ('mic' in first)
        assert ('operating_mic' in first)
        assert ('acronym' in first)
        assert ('name' in first)
        assert ('city' in first)
        assert ('country' in first)
        assert ('timezone' in first)
        assert ('website' in first)

    def test_get_market_today_hours(self):
        market = r.get_market_today_hours(self.nyse)
        assert ('date' in market)
        assert ('is_open' in market)
        assert ('opens_at' in market)
        assert ('closes_at' in market)
        assert ('extended_opens_at' in market)
        assert ('extended_closes_at' in market)
        assert ('previous_open_hours' in market)
        assert ('next_open_hours' in market)

    def test_get_market_next_open_hours(self):
        market = r.get_market_next_open_hours(self.amex)
        assert ('date' in market)
        assert ('is_open' in market)
        assert ('opens_at' in market)
        assert ('closes_at' in market)
        assert ('extended_opens_at' in market)
        assert ('extended_closes_at' in market)
        assert ('previous_open_hours' in market)
        assert ('next_open_hours' in market)

    def test_get_market_next_open_hours_after_date(self):
        market = r.get_market_next_open_hours_after_date(self.nasdaq, self.today)
        assert ('date' in market)
        assert ('is_open' in market)
        assert ('opens_at' in market)
        assert ('closes_at' in market)
        assert ('extended_opens_at' in market)
        assert ('extended_closes_at' in market)
        assert ('previous_open_hours' in market)
        assert ('next_open_hours' in market)

    def test_get_market_hours(self):
        market = r.get_market_hours(self.nasdaq, self.today)
        assert ('date' in market)
        assert ('is_open' in market)
        assert ('opens_at' in market)
        assert ('closes_at' in market)
        assert ('extended_opens_at' in market)
        assert ('extended_closes_at' in market)
        assert ('previous_open_hours' in market)
        assert ('next_open_hours' in market)

    def test_currency_pairs(self):
        currency = r.get_currency_pairs()
        assert currency
        first = currency[0]
        assert ('asset_currency' in first)
        assert ('display_only' in first)
        assert ('id' in first)
        assert ('max_order_size' in first)
        assert ('min_order_size' in first)
        assert ('min_order_price_increment' in first)
        assert ('min_order_quantity_increment' in first)
        assert ('name' in first)
        assert ('quote_currency' in first)
        assert ('symbol' in first)
        assert ('tradability' in first)

    @pytest.mark.xfail()
    def test_market_fail(self):
        market = r.get_market_hours(self.fake, self.today)
        assert market

    @pytest.mark.xfail()
    def test_market_date_fail(self):
        market = r.get_market_hours(self.nasdaq, self.american_time)
        assert market

class TestProfiles:
    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)

    @classmethod
    def teardown_class(cls):
        r.logout()

    def test_load_account_profile(self):
        profile = r.load_account_profile(info=None)
        assert profile
        assert ('url' in profile)
        assert ('portfolio_cash' in profile)
        assert ('can_downgrade_to_cash' in profile)
        assert ('user' in profile)
        assert ('account_number' in profile)
        assert ('type' in profile)
        assert ('created_at' in profile)
        assert ('updated_at' in profile)
        assert ('deactivated' in profile)
        assert ('deposit_halted' in profile)
        assert ('only_position_closing_trades' in profile)
        assert ('buying_power' in profile)
        assert ('cash_available_for_withdrawal' in profile)
        assert ('cash' in profile)
        assert ('cash_held_for_orders' in profile)
        assert ('uncleared_deposits' in profile)
        assert ('sma' in profile)
        assert ('sma_held_for_orders' in profile)
        assert ('unsettled_funds' in profile)
        assert ('unsettled_debit' in profile)
        assert ('crypto_buying_power' in profile)
        assert ('max_ach_early_access_amount' in profile)
        assert ('cash_balances' in profile)
        assert ('margin_balances' in profile)
        assert ('sweep_enabled' in profile)
        assert ('instant_eligibility' in profile)
        assert ('option_level' in profile)
        assert ('is_pinnacle_account' in profile)
        assert ('rhs_account_number' in profile)
        assert ('state' in profile)
        assert ('active_subscription_id' in profile)
        assert ('locked' in profile)
        assert ('permanently_deactivated' in profile)
        assert ('received_ach_debit_locked' in profile)
        assert ('drip_enabled' in profile)
        assert ('eligible_for_fractionals' in profile)
        assert ('eligible_for_drip' in profile)
        assert ('eligible_for_cash_management' in profile)
        assert ('cash_management_enabled' in profile)
        assert ('option_trading_on_expiration_enabled' in profile)
        assert ('cash_held_for_options_collateral' in profile)
        assert ('fractional_position_closing_only' in profile)
        assert ('user_id' in profile)

    def test_basic_profile(self):
        profile = r.load_basic_profile(info=None)
        assert profile
        assert ('user' in profile)
        assert ('address' in profile)
        assert ('city' in profile)
        assert ('state' in profile)
        assert ('zipcode' in profile)
        assert ('phone_number' in profile)
        assert ('marital_status' in profile)
        assert ('date_of_birth' in profile)
        assert ('citizenship' in profile)
        assert ('country_of_residence' in profile)
        assert ('number_dependents' in profile)
        assert ('signup_as_rhs' in profile)
        assert ('tax_id_ssn' in profile)
        assert ('updated_at' in profile)

    def test_investment_profile(self):
        profile = r.load_investment_profile(info=None)
        assert profile
        assert ('user' in profile)
        assert ('total_net_worth' in profile)
        assert ('annual_income' in profile)
        assert ('source_of_funds' in profile)
        assert ('investment_objective' in profile)
        assert ('investment_experience' in profile)
        assert ('liquid_net_worth' in profile)
        assert ('risk_tolerance' in profile)
        assert ('tax_bracket' in profile)
        assert ('time_horizon' in profile)
        assert ('liquidity_needs' in profile)
        assert ('investment_experience_collected' in profile)
        assert ('suitability_verified' in profile)
        assert ('option_trading_experience' in profile)
        assert ('professional_trader' in profile)
        assert ('understand_option_spreads' in profile)
        assert ('interested_in_options' in profile)
        assert ('updated_at' in profile)

    def test_portfolio_profile(self):
        profile = r.load_portfolio_profile(info=None)
        assert profile
        assert ('url' in profile)
        assert ('account' in profile)
        assert ('start_date' in profile)
        assert ('market_value' in profile)
        assert ('equity' in profile)
        assert ('extended_hours_market_value' in profile)
        assert ('extended_hours_equity' in profile)
        assert ('extended_hours_portfolio_equity' in profile)
        assert ('last_core_market_value' in profile)
        assert ('last_core_equity' in profile)
        assert ('last_core_portfolio_equity' in profile)
        assert ('excess_margin' in profile)
        assert ('excess_maintenance' in profile)
        assert ('excess_margin_with_uncleared_deposits' in profile)
        assert ('excess_maintenance_with_uncleared_deposits' in profile)
        assert ('equity_previous_close' in profile)
        assert ('portfolio_equity_previous_close' in profile)
        assert ('adjusted_equity_previous_close' in profile)
        assert ('adjusted_portfolio_equity_previous_close' in profile)
        assert ('withdrawable_amount' in profile)
        assert ('unwithdrawable_deposits' in profile)
        assert ('unwithdrawable_grants' in profile)

    def test_security_profile(self):
        profile = r.load_security_profile(info=None)
        assert profile
        assert ('user' in profile)
        assert ('object_to_disclosure' in profile)
        assert ('sweep_consent' in profile)
        assert ('control_person' in profile)
        assert ('control_person_security_symbol' in profile)
        assert ('security_affiliated_employee' in profile)
        assert ('security_affiliated_firm_relationship' in profile)
        assert ('security_affiliated_firm_name' in profile)
        assert ('security_affiliated_person_name' in profile)
        assert ('security_affiliated_address' in profile)
        assert ('security_affiliated_address_subject' in profile)
        assert ('security_affiliated_requires_duplicates' in profile)
        assert ('stock_loan_consent_status' in profile)
        assert ('agreed_to_rhs' in profile)
        assert ('agreed_to_rhs_margin' in profile)
        assert ('rhs_stock_loan_consent_status' in profile)
        assert ('updated_at' in profile)

    def test_user_profile(self):
        profile = r.load_user_profile(info=None)
        assert profile
        assert ('url' in profile)
        assert ('id' in profile)
        assert ('id_info' in profile)
        assert ('username' in profile)
        assert ('email' in profile)
        assert ('email_verified' in profile)
        assert ('first_name' in profile)
        assert ('last_name' in profile)
        assert ('origin' in profile)
        assert ('profile_name' in profile)
        assert ('created_at' in profile)

    def test_crypto_profile(self):
        profile = r.load_crypto_profile(info=None)
        assert profile
        assert ('apex_account_number' in profile)
        assert ('created_at' in profile)
        assert ('id' in profile)
        assert ('rhs_account_number' in profile)
        assert ('status' in profile)
        assert ('status_reason_code' in profile)
        assert ('updated_at' in profile)
        assert ('user_id' in profile)

    @pytest.mark.xfail()
    def test_key_failed(self):
        profile = r.load_account_profile(info='cheese')
        assert profile

    @pytest.mark.xfail()
    def test_login_failed(self):
        r.logout()
        profile = r.load_account_profile(info=None)
        assert profile

class TestOrders:
    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)
    
    def test_find_stock_orders(cls):
        def isFloat(f):
            try:
                float(f)
                return True
            except ValueError:
                return False

        orderHistory = r.find_stock_orders()
        assert orderHistory

        for order in orderHistory:
            assert isFloat(order['quantity'])
            assert isFloat(order['cumulative_quantity'])
            if(order['state'] == 'filled'):
                assert (order['quantity'] == order['cumulative_quantity'])
                
class TestAccountInformation:
    @classmethod
    def setup_class(cls):
        totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
        login = r.login(os.environ['robin_username'], os.environ['robin_password'], mfa_code=totp)
    
    def test_get_stock_loan_payments(cls):
        def isFloat(f):
            try:
                float(f)
                return True
            except ValueError:
                return False

        loanPayments = r.get_stock_loan_payments()
        assert loanPayments
        for payment in loanPayments:
            assert ('amount' in payment)
            assert isFloat(payment['amount']['amount'])
            assert ('symbol' in payment)
            assert ('description' in payment)
        
    def test_get_interest_payments(cls):
        def isFloat(f):
            try:
                float(f)
                return True
            except ValueError:
                return False
            
        interests = r.get_interest_payments()
        assert (interests)
        for interest in interests:
            assert ('amount' in interest)
            assert isFloat(interest['amount']['amount'])
            assert ('direction' in interest)
            assert ('payout_type' in interest)
            assert ('reason' in interest)


from robin_stocks.tda.helper import format_inputs, login_required, request_get
from robin_stocks.tda.urls import URLS


@login_required
@format_inputs
def get_quote(ticker, jsonify=None):
    """ Gets quote information for a single stock.

    :param ticker: The ticker of the stock.
    :type ticker: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.quote(ticker)
    data, error = request_get(url, None, jsonify)
    return data, error


@login_required
@format_inputs
def get_quotes(tickers, jsonify=None):
    """ Gets quote information for multiple stocks. The stock string should be comma separated with no spaces.

    :param ticker: The string list of stock tickers.
    :type ticker: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.quotes()
    payload = {
        "symbol": tickers
    }
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_price_history(ticker, period_type, frequency_type, frequency,
                      period=None, start_date=None, end_date=None, needExtendedHoursData=True, jsonify=None):
    """ Gets the price history of a stock.

    :param ticker: The stock ticker.
    :type ticker: str
    :param period_type: The type of period to show. Valid values are day, month, year, or ytd (year to date). Default is day.
    :type period_type: str
    :param frequency_type: The type of frequency with which a new candle is formed. \
        Valid frequencyTypes by period_type (defaults marked with an asterisk):\n
        * day: minute*
        * month: daily, weekly* 
        * year: daily, weekly, monthly* 
        * ytd: daily, weekly*
    :type frequency_type: str
    :param frequency: The number of the frequencyType to be included in each candle. \
        Valid frequencies by frequencyType (defaults marked with an asterisk):\n
        * minute: 1*, 5, 10, 15, 30
        * daily: 1*
        * weekly: 1*
        * monthly: 1*
    :type frequency: str
    :param period: The number of periods to show. Valid periods by periodType (defaults marked with an asterisk):\n
        * day: 1, 2, 3, 4, 5, 10*
        * month: 1*, 2, 3, 6
        * year: 1*, 2, 3, 5, 10, 15, 20
        * ytd: 1*
    :type period: Optional[str]
    :param start_date: Start date as milliseconds since epoch. If startDate and endDate are provided, period should not be provided.
    :type start_date: Optional[str]
    :param end_date: End date as milliseconds since epoch. If startDate and endDate are provided, period should not be provided. Default is previous trading day.
    :type end_date: Optional[str]
    :param needExtendedHoursData: true to return extended hours data, false for regular market hours only. Default is true.
    :type needExtendedHoursData: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    if (start_date or end_date) and period:
        raise ValueError(
            "If start_date and end_date are provided, period should not be provided.")
    url = URLS.price_history(ticker)
    payload = {
        "periodType": period_type,
        "frequencyType": frequency_type,
        "frequency": frequency,
        "needExtendedHoursData": needExtendedHoursData
    }
    if period:
        payload["period"] = period
    if start_date:
        payload["startDate"] = start_date
    if end_date:
        payload["endDate"] = end_date
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def search_instruments(ticker_string, projection, jsonify=None):
    """ Gets a list of all the instruments data for tickers that match a search string.

    :param ticker_string: Value to pass to the search. See projection description for more information.
    :type ticker_string: str
    :param projection: The type of request:\n
        * symbol-search: Retrieve instrument data of a specific symbol or cusip
        * symbol-regex: Retrieve instrument data for all symbols matching regex. Example: symbol=XYZ.* will return all symbols beginning with XYZ
        * desc-search: Retrieve instrument data for instruments whose description contains the word supplied. Example: symbol=FakeCompany will return all instruments with FakeCompany in the description.
        * desc-regex: Search description with full regex support. Example: symbol=XYZ.[A-C] returns all instruments whose descriptions contain a word beginning with XYZ followed by a character A through C.
        * fundamental: Returns fundamental data for a single instrument specified by exact symbol.
    :type projection: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.instruments()
    payload = {
        "symbol": ticker_string,
        "projection": projection
    }
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_instrument(cusip, jsonify=None):
    """ Gets instrument data for a specific stock.

    :param cusip: The CUSIP for a stock.
    :type cusip: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.instrument(cusip)
    data, error = request_get(url, None, jsonify)
    return data, error


@login_required
@format_inputs
def get_option_chains(ticker, contract_type="ALL", strike_count="10", include_quotes="FALSE", strategy="SINGLE", interval=None, strike_price=None,
                      range_value="ALL", from_date=None, to_date=None, volatility=None, underlying_price=None, interest_rate=None, 
                      days_to_expiration=None, exp_month="ALL", option_type="ALL", jsonify=None):
    """ Gets instrument data for a specific stock.

    :param ticker: The stock ticker.
    :type ticker: str
    :param contract_type: Type of contracts to return in the chain. Can be CALL, PUT, or ALL. Default is ALL.
    :type contract_type: Optional[str]
    :param strike_count: The number of strikes to return above and below the at-the-money price.
    :type strike_count: Optional[str]
    :param include_quotes: Include quotes for options in the option chain. Can be TRUE or FALSE. Default is FALSE.
    :type include_quotes: Optional[str]
    :param strategy: Passing a value returns a Strategy Chain. Possible values are SINGLE, ANALYTICAL (allows use of the volatility, \
        underlyingPrice, interestRate, and daysToExpiration params to calculate theoretical values), COVERED, VERTICAL, CALENDAR, \
        STRANGLE, STRADDLE, BUTTERFLY, CONDOR, DIAGONAL, COLLAR, or ROLL. Default is SINGLE.
    :type strategy: Optional[str]
    :param interval: Strike interval for spread strategy chains (see strategy param).
    :type interval: Optional[str]
    :param strike_price: Provide a strike price to return options only at that strike price.
    :type strike_price: Optional[str]
    :param range_value: Returns options for the given range. Default is ALL. Possible values are:\n
        * ITM: In-the-money
        * NTM: Near-the-money
        * OTM: Out-of-the-money
        * SAK: Strikes Above Market
        * SBK: Strikes Below Market
        * SNK: Strikes Near Market
        * ALL: All Strikes
    :type range_value: Optional[str]
    :param from_date: Only return expirations after this date. For strategies, expiration refers to the \
        nearest term expiration in the strategy. Valid ISO-8601 formats are: yyyy-MM-dd and yyyy-MM-dd'T'HH:mm:ssz.
    :type from_date: Optional[str]
    :param to_date: Only return expirations before this date. For strategies, expiration refers to the \
        nearest term expiration in the strategy. Valid ISO-8601 formats are: yyyy-MM-dd and yyyy-MM-dd'T'HH:mm:ssz.
    :type to_date: Optional[str]
    :param volatility: Volatility to use in calculations. Applies only to ANALYTICAL strategy chains (see strategy param).
    :type volatility: Optional[str]
    :param underlying_price: Underlying price to use in calculations. Applies only to ANALYTICAL strategy chains (see strategy param).
    :type underlying_price: Optional[str]
    :param interest_rate: Interest rate to use in calculations. Applies only to ANALYTICAL strategy chains (see strategy param).
    :type interest_rate: Optional[str]
    :param days_to_expiration: Days to expiration to use in calculations. Applies only to ANALYTICAL strategy chains (see strategy param).
    :type days_to_expiration: Optional[str]
    :param exp_month: Return only options expiring in the specified month. Month is given in the three character format. Example: JAN. Default is ALL.
    :type exp_month: Optional[str]
    :param option_type: Type of contracts to return. Default is ALL. Possible values are:\n
        * S: Standard contracts
        * NS: Non-standard contracts
        * ALL: All contracts
    :type option_type: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.option_chains()
    payload = {
        "symbol": ticker,
        "contractType": contract_type,
        "strikeCount": strike_count,
        "includeQuotes": include_quotes,
        "strategy": strategy,
        "range": range_value,
        "expMonth": exp_month,
        "optionType": option_type
    }
    if interval:
        payload["interval"] = interval
    if strike_price:
        payload["strike"] = strike_price
    if from_date:
        payload["fromDate"] = from_date
    if to_date: 
        payload["toDate"] = to_date
    if volatility:
        payload["volatility"] = volatility
    if underlying_price:
        payload["underlyingPrice"] = underlying_price
    if interest_rate:
        payload["interestRate"] = interest_rate
    if days_to_expiration:
        payload["daysToExpiration"] = days_to_expiration
    data, error = request_get(url, payload, jsonify)
    return data, error



"""Holds the session header and other global variables."""
from requests import Session

DATA_DIR_NAME = ".tokens"
PICKLE_NAME = "tda.pickle"
RETURN_PARSED_JSON_RESPONSE = False # Flag on whether to automatically parse request responses.
LOGGED_IN = False  # Flag on whether or not the user is logged in.

# The session object for making get and post requests.
SESSION = Session()
SESSION.headers = {
    "Accept": "*/*",
    "Accept-Encoding": "gzip",
    "Accept-Language": "en-US",
    "Host": "api.tdameritrade.com",
    "Connection": "keep-alive",
    "Content-Type": "application/json;charset=UTF-8",
    "User-Agent": "*",
    "sec-ch-ua-mobile":"?0",
    "Sec-Fetch-Dest":"empty",
    "Sec-Fetch-Mode":"cors",
    "Sec-Fetch-Site":"same-site"
}



from uuid import uuid4
import datetime
import sys
import pyotp
import os
from dotenv import load_dotenv
load_dotenv()
import requests

import robin_stocks.robinhood as r
import robin_stocks.gemini as g
import robin_stocks.tda as t

single_stock = 'AAPL'

totp  = pyotp.TOTP(os.environ['robin_mfa']).now()
print("Current OTP:", totp)
login = r.login(os.environ['robin_username'], os.environ['robin_password'], store_session=True, mfa_code=totp)
print("login data is ", login)
print(r.load_account_profile(info=None))
exit()
# print("===")
# print("running test at {}".format(datetime.datetime.now()))
# print("===")

# order = r.get_watchlist_by_name("test")
# print(order)

# t.login_first_time("2HURJS3NJN4BCXZ84HHHDACAGRTCNDDU",
# "BpjW+hnsy9tO5OFBv0QUIpCLvTx4/2sLr1xJD3uao8PL9/qCuFQ6g5IchzHBTY6Az8nf7CrO5klTrcF8tpRH/l6e7sTQCluzDfUjqVJJC4Vmx2S1XOsbYKpm+jFiIg84+H0f/pUrrgtFiy3bKMUzz/M6TTsAwN1VWeaet7aoMsguvnsq7nNXCkR2COlKbFAlFvg69YvagiJFC0wdBwhLvYPLAH6cxIGTT6CYMMIZ06qsqCQ7jzVLqMd24pw2FpVtzQYgzb8XVORQ3KWrtrX7tiecsSOVZxNKz0zc30rZPeEecwWgZSMLF2KZseQs8unl8Kv6GRV397FYUskduy9/Whk8acP9EfSHvG4O8zAhXr0B13dRHdSMcVqXgu72FX5PzuAj7B98MT0ew1OdAP3Qw3/bPE0Wg3iRrFGS2XeBAc1lM26MgXcp0sgyOXykAwpepvp4d+HKAZ2fqZTSk/XQlLpfSNzG/S+yVkp6lQiTCZRPUavCk7+MiqD/uSHYBUP+suNlkdIy6CJDk4VJk2lKjdXDadn2qny509cmn100MQuG4LYrgoVi/JHHvlXs39mVzQtKjpZ+4SdOckKKkzf+5YhvTSbPDJhijmMOaGKKKh0ziBaU5h8WI8UiKbBWnxPReHQoSYDd/W6ZR7y4ngBnZSeIMc0izFfzwZFcSK0mAruUcQviqReFNj4RPMluq2sY9BBEwf6KdO4IthWEihE9imCzU8ZdSalYwJ6VRDN7A2G1gMW1SB7g/rF7UwqmHpTxdlBYjoUD4ypDUgpWuWpQ80FCNor4GFdYAQeoI0BPVAAtHNg5IcqCfy71XB8WhL0GKuuTrL5UjBxZqA3iDBPnL+NpoY4FZppkD6noqF/b++hPnd0t0b0kYJe09zh/cC2OtOoTWBoYTDdB/jNRLCd+PNrD6Q/rFOx/Wx1mujS1YjlQqwHA5G1cd02ODUM1OO04aBy48eTuQXupWNDFsLqWqwauw6ZBPqszQuNCKs0v+nebBcBBWhlW9ULD8pDr9qMcpR6iYMbxvQA8ho7Bc77+FB2Nh/ajEqL84sRDsuy71tIpelcYgQ4yOoHB6ebNHkFJtOwTuX4gnGF8DN8KrhQ7OwVGIPtf1yOK212FD3x19z9sWBHDJACbC00B75E",
# "G2aUpN+qXx61tNumt4aKThC7W0u8IvWmv1W7rSawftRxkskYUPgpVJSCZ4tmgXHmLlZu8r+uTIT5uBLMqN5peW8n4dAatdvLLLt7TFqcw63m7fCU6qaiWxhzaLF3QfPq7h3DPTCNVr+/SQmKOmqBi+t6KYyXGi94jTiQOB1BjkTmE9qPf35AL9cr5kKYbVXJEOeTNsedcFMXjo8+UCPLrtnW2rCFa1FOjmjMCxZX4M/RuwXVa/5+tTXq3+jOlBmabjz7SiDP42ooPVkhncBeXP4oq8zYnBnhLnLxx0falDcbFY1K8QZ6tftHsSFCjF9QkzigQpbyhAjaqfm7C+46+w/89I15eGJn9HsWyYNd2Y6f8LJg8bMoRtnQoqshwUEtSiJqfd2vq4TaabLn6snBcuBoj2m6qyDSfM6oQA/vM8PIYqzfmsI7gok6jWf100MQuG4LYrgoVi/JHHvl2iAD+MiwRt1Jcok/7oEYYXXBmA1DiKQSi8bRcOG2zSogi5JMtUiqV9YrHrZ0V0U2YwkxRT4JlUJPj/kU+09YUOn5qxnog5CvI4eGyndWLq7PRa3TJ7bEpEuMTW765C91Ei2SKKqBDfzpzjDkX6whFvH4pB7cP4SxviNx6nmCB8Ywd1Qh0MqCp3251b+mRvbHW24ZbdPPJZsod0OomFEpCJ/l/G4Cb0bnJ0BMarONjR5J+Phk6xVzUrCaFI7+ho8IL9KGsJxZI7WaE1IHnpeJBaJMxqgSUa/6bOQgmuIUuOMdycCNVv59XO1tg7PvKAGJ3eGh3pmojy89Q28bhBpE9aJfAKqlke8HD26akrOgBZVD85ch13yTBCfuHPSkxZ9iy+XeYq78EU4WE7qV4ZfUjtzB2ZJOZzR/FRe0NpGuiRe0k1MOCnTsh5SgwNs=212FD3x19z9sWBHDJACbC00B75E")

# t.login_first_time("2HURJS3NJN4BCXZ84HHHDACAGRTCNDDU",
#                    "9NfI4Or1pbhSmfCNg9Szw9uH7z46cCQzFn83ND7vehm7dmvJhEJIu7eK9PvpF5ymUc9OVlUTJYXZocSYMQYOgEPdYTYeQN4K8Wg4PFK7TWDA3Y7s3gRcjbQubfOL5aRDpR9l2bS/3nDQvFOQTHS/8wJg7lqVJMBHWH7w5HfIa3uMVrtg6bVgjDu0yMOzv/jnyKmn+ks95HFAj45ccM9IpgI8cvkwlOw72A2vnJAJmuITs8a4nAwNKPnO4yuApsUR92G3LDdAK+eBTcN7pBeEh7eZU49Mu1291wAY2WthWWl26qh8qGuAwmsr5RB+PZbo4+ebISVZetLMej9yFhhWBnB/cRD+mI9RGBRMtFL5WXCKVH3OHT41xmcHa1+5f3Hz5CmBAmGpJH+i3roWPSSbDOngFyrLcmu0fu+9hDmRDdL/tk+PU4qSBG8wTZbsNZhwKokZrOXa6NLOUAC7kMBupH8iiDhKU1gVjhbiPSXGcd9Z22OpmzhBT5SyftZFs1ysz/YCoNCUG0XpRe6IrUVDiGnAlwuuhH/TcneFmoczHi1NO10N0100MQuG4LYrgoVi/JHHvlKEUe2uBmLFCkJskrle34M3SQixpz4bUKmZoAsFwQAurNIGf9w+MjnNxOERKw+hO8nNbGKDvJ2BYZ+8NiuahbywpD0PqvWFrmmu4sQ6NrXT/pj4QfKqcfXqUhpn1Eb2jERC12qGXE9+rIzUiKX0LiZmLsjP1xYPwPSh/40DMYaGXXjYHDX9EJg4bwZgfguQuwquDWAw3jXUjGhAnK5aLo3Y7G499Gu/X8oWkB+HprlOJddjNHg6lMvhJKRIO7mHQSpLaCrqbiK9jgd9QdJ1gAUS96EmaXwbmnpiJfS6cs67XgoWGoBsWAMCKv+MZTt0tK8+xzPEB6gyeGFdxtrxE1X0wdl6gTncIxghPTCREzmNdqlQw1AdPFcVh9x8WIy2QKvxj7RDK5/6bvDm0I0GgPQr46NBN87bG+7u+orwcs6LrTOTA1NY6dlgKNQtKlf9npQcrRAEk2B0DG9mzolLN7Za203yEXfl4OzcbyjNUmxTb39bdMqyO3csEKmp8I89nPFlNhVMVJLcOCTXK3Bi5fC87ZyccCPcSuvw7j1YRsntMM1vkQ==212FD3x19z9sWBHDJACbC00B75E",
#                    "HrE6Hs+9rKxQczwG61se9jjokcelBTuq7rG1+YePxJfnyOlopkcku2fpCfBwB4bJ8++mpxDRz32InTMcWhnMzV0h6MFS3GXdO0Cwy3NZuslC9hUhAVtviGNUbSP5fS0lpnZy6w63S9ygIWor6Y4EKdIGirdLhk4L1pSV30i0qd6zYSFy2prXPGILRUC+LFBIK2xQ4pc0r47b/qSDeickETtXM7mdBNlb+9DboQ9aaf0AOFF6quTHpcZOPeryjWM1/gv7LrmwHeyTvixfdO53tjiunFGfTRAMIaEYWlYjLwTUWDfFkTT5lUOEOAupQ+VfP4dVCuqKwgN+z/X1QQDaWWl40PmymsZcUaIJirpK3BlWlI8EuBUorRKrG2Df51C0m+m7zWpy2/2me5NGQUm7SN1fSytErl8EVLWCGUcWcmBzeV+7O610vHqCwNs100MQuG4LYrgoVi/JHHvl6JfdrQEBiAEFWXP+RSVEgq2cJnqsyTJPV+4pwLUzkOKwp46ZpL3ckQthosey7EOstBhr461mDinfXNcnQ5d5CYH7C+aBrcQ0UPFsdh8RquavV3hhuaK3JVufmiy7NTfKz198mWSlMOluz0kwdPEh66c4Bds374Zwi/pgLnoSIZvkLBT1iAxkCTtptigwdF0WWuu4gebHgVZO8Tx7VOw+9IcRHUb+QjAlEl52uBJDkkVzk73O1RlVlpzGfXRzYvzkcgOkgzgbSi8G2OAavUUwUC1NRIWFR2rJZGhlnaI8QcP138kWl87DXfEx4UJa56YS6Cxhf2YKl5nug6gta7VlqZhwaIX0BM/5omLSrVoMvDI/304/507lxRk0w10xTQOInIb3WCbWVe9ShU5uNW2K8eUwZpp9ECQuF88zDvX7dUp5uG/wRgWWe1qdqro=212FD3x19z9sWBHDJACbC00B75E",
#                    "w_Pt4oNsijRB_b5LdvjBZsM3uPE8uB0503lQk8kf7YE=")

# "G2aUpN%2BqXx61tNumt4aKThC7W0u8IvWmv1W7rSawftRxkskYUPgpVJSCZ4tmgXHmLlZu8r%2BuTIT5uBLMqN5peW8n4dAatdvLLLt7TFqcw63m7fCU6qaiWxhzaLF3QfPq7h3DPTCNVr%2B%2FSQmKOmqBi%2Bt6KYyXGi94jTiQOB1BjkTmE9qPf35AL9cr5kKYbVXJEOeTNsedcFMXjo8%2BUCPLrtnW2rCFa1FOjmjMCxZX4M%2FRuwXVa%2F5%2BtTXq3%2BjOlBmabjz7SiDP42ooPVkhncBeXP4oq8zYnBnhLnLxx0falDcbFY1K8QZ6tftHsSFCjF9QkzigQpbyhAjaqfm7C%2B46%2Bw%2F89I15eGJn9HsWyYNd2Y6f8LJg8bMoRtnQoqshwUEtSiJqfd2vq4TaabLn6snBcuBoj2m6qyDSfM6oQA%2FvM8PIYqzfmsI7gok6jWf100MQuG4LYrgoVi%2FJHHvl2iAD%2BMiwRt1Jcok%2F7oEYYXXBmA1DiKQSi8bRcOG2zSogi5JMtUiqV9YrHrZ0V0U2YwkxRT4JlUJPj%2FkU%2B09YUOn5qxnog5CvI4eGyndWLq7PRa3TJ7bEpEuMTW765C91Ei2SKKqBDfzpzjDkX6whFvH4pB7cP4SxviNx6nmCB8Ywd1Qh0MqCp3251b%2BmRvbHW24ZbdPPJZsod0OomFEpCJ%2Fl%2FG4Cb0bnJ0BMarONjR5J%2BPhk6xVzUrCaFI7%2Bho8IL9KGsJxZI7WaE1IHnpeJBaJMxqgSUa%2F6bOQgmuIUuOMdycCNVv59XO1tg7PvKAGJ3eGh3pmojy89Q28bhBpE9aJfAKqlke8HD26akrOgBZVD85ch13yTBCfuHPSkxZ9iy%2BXeYq78EU4WE7qV4ZfUjtzB2ZJOZzR%2FFRe0NpGuiRe0k1MOCnTsh5SgwNs%3D212FD3x19z9sWBHDJACbC00B75E"

# blah = t.generate_encryption_passcode()
# print(blah)
# print(type(blah))
# print(blah)
# print(blah.decode())
# print(type(blah.decode()))
print(t.login("w_Pt4oNsijRB_b5LdvjBZsM3uPE8uB0503lQk8kf7YE="))
# data, _ = t.get_account("233798512","orders", jsonify=False)
# print(type(data) == requests.models.Response)
# print(data)
order = {
    "orderType": "MARKET",
    "session": "NORMAL",
    "duration": "DAY",
    "orderStrategyType": "SINGLE",
    "orderLegCollection": [
        {
            "instruction": "Buy",
            "quantity": 1,
            "instrument": {
                "symbol": "AMC",
                "assetType": "EQUITY"
            }
        }
    ]
}
# from re import IGNORECASE, match, split
# data, err = t.place_order("233798512", order, False)
# print(data)
# print(t.get_order_number(data))
# data, err = t.get_instrument("88160R101", True)
# data, err = t.get_hours_for_market("OPTION", "2021-02-16", True)
# data, err = t.get_option_chains("AAPL", "PUT", )
data, err = t.get_option_chains("AAPL")
print(data, err)
exit()
# curl -X POST --header "Content-Type: application/x-www-form-urlencoded" -d "grant_type=refresh_token&refresh_token=qjJBft4oazSHjgTeUEt95qJlQJlYxYqWv7Ju62OTE73hDZcTeyI01GlbSJa2eiQjG1o9guz8CatR6NugyJMOIy08qBMYB3EapZVddTkA3pMyztgsBym36RTMPx9XA9B%2BUlTBjqRqQoThVCH5v1PoR6sbiKzqjiWznt3MmQfkPHM3%2FsDXBzueknatunvFNa7JJ1DQxC5OFUUuZnkN6MQpupNfepeymd6W1GlatQW1lJyLaoGJrvnKbY%2BxleicGNwOQI6iRZ30djWgKdyfeBvXEZp7ABTFf2%2FczVnXGa43ow4UjmJKwcpqKhR1qLHe1KN7BgwnvXmlp9bi7TzXPCYDxix5nGqmwaXPx9ZxqraKXcFYpXj5RC5OHSdnr0%2B6RgbrPoEV%2FS0H%2B8sTxmrRYeOEOgF%2B5%2F4uP4M2MvYkm5%2FntkX0xEXmiLyLQaR%2Bc%2Bj100MQuG4LYrgoVi%2FJHHvlKZgyoTHJmrFJcCQFbf4YdWPL9SLHEuRgJ3mvqJenT41nllMAhCAMhPJnXfG1hFV0LkDZoKIY%2FZwXP3zQndWX1cwQBiQlD8Xmzjb9B%2F%2BXY5uWtxhXcRhpZvc%2F2Vwu2qcXZ2NTL%2FqpSJhqK4J8olnLS398E1TqJpkw0RzoM1jajMY9yGhN3ftzVslaHGAMUZtUJcvnKfE6Osq5FzusUINnCjIdhBZhwknkzvja8O2D34a%2BTnQe3c1YZR%2BTYy3dek4%2B1Z%2FZ6qb%2FWpX304aziUoF4Id4QkvDyFgL0wM81kmMYOJd%2B2VfLUjkcGiSPRCUizpUhTUbSBgfATnnLipAxiqBPi6W%2B22zQNQUGUv4rZzes9IivkvYHzkj%2FTOdojFNdGIrs1POjpqH%2BvUxV0mxZVwVn3Bx27hjQd31WL%2BP8zjSYAeAKAeOkDAHQkOg8lk%3D212FD3x19z9sWBHDJACbC00B75E&client_id=2HURJS3NJN4BCXZ84HHHDACAGRTCNDDU" "https://api.tdameritrade.com/v1/oauth2/token"

from cryptography.fernet import Fernet

key = Fernet.generate_key()
cipher_suite = Fernet(key)
cipher_text = {"food": cipher_suite.encrypt("cheese".encode())}

print("the cipher text is ", cipher_text)
plain_text = cipher_suite.decrypt(cipher_text["food"])
print("plain text is ", plain_text.decode())

exit()

import hashlib
import hmac
import pickle
from pathlib import Path

my_data = {"food": "cheese", "number": 1}

def make_digest(message):
    return hmac.new(b'12345', message, hashlib.sha1).hexdigest()


pickled_data = pickle.dumps(my_data)
digest = make_digest(pickled_data)
my_path = Path(".").joinpath("test.pickle")
print("my digest is ", digest)
header = '%s %s' % (digest, len(pickled_data))
blah = header + '\n'
with my_path.open("wb") as pickle_file:
    pickle_file.write(blah.encode())
    pickle_file.write(pickled_data)

with my_path.open("rb") as pickle_file:
    first_line = pickle_file.readline()
    print(first_line)
    incoming_digest, incoming_length = first_line.decode().split(' ')
    incoming_length = int(incoming_length)

    print ('\nREAD:', incoming_digest, incoming_length)

    incoming_pickled_data = pickle_file.read(incoming_length)

    actual_digest = make_digest(incoming_pickled_data)
    print ('ACTUAL:', actual_digest)

    pickle_data = pickle.loads(incoming_pickled_data)
    print(pickled_data["food"])

# test = make_digest("cheese".encode())
# print("test is ", test)
# print(hmac.digest(b'12345', test , hashlib.sha1))



"""Contains functions for getting all the information tied to a user account."""
from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.urls import *


@login_required
def load_account_profile(account_number=None, info=None, dataType="indexzero"):
    """Gets the information associated with the accounts profile,including day
    trading information and cash being held by Robinhood.

    :param acccount_number: the robinhood account number.
    :type acccount_number: Optional[str]
    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :param dataType: Determines how to filter the data. 'regular' returns the unfiltered data. \
    'results' will return data['results']. 'pagination' will return data['results'] and append it with any \
    data that is in data['next']. 'indexzero' will return data['results'][0].
    :type dataType: Optional[str]
    :returns: The function returns a dictionary of key/value pairs. \
    If a string is passed in to the info parameter, then the function will return \
    a string corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * url
                      * portfolio_cash
                      * can_downgrade_to_cash
                      * user
                      * account_number
                      * type
                      * created_at
                      * updated_at
                      * deactivated
                      * deposit_halted
                      * only_position_closing_trades
                      * buying_power
                      * cash_available_for_withdrawal
                      * cash
                      * cash_held_for_orders
                      * uncleared_deposits
                      * sma
                      * sma_held_for_orders
                      * unsettled_funds
                      * unsettled_debit
                      * crypto_buying_power
                      * max_ach_early_access_amount
                      * cash_balances
                      * margin_balances
                      * sweep_enabled
                      * instant_eligibility
                      * option_level
                      * is_pinnacle_account
                      * rhs_account_number
                      * state
                      * active_subscription_id
                      * locked
                      * permanently_deactivated
                      * received_ach_debit_locked
                      * drip_enabled
                      * eligible_for_fractionals
                      * eligible_for_drip
                      * eligible_for_cash_management
                      * cash_management_enabled
                      * option_trading_on_expiration_enabled
                      * cash_held_for_options_collateral
                      * fractional_position_closing_only
                      * user_id
                      * rhs_stock_loan_consent_status

    """
    url = account_profile_url(account_number)
    if account_number is not None:
         data = request_get(url)
    else:
        data = request_get(url, dataType)
    return(filter_data(data, info))


@login_required
def load_basic_profile(info=None):
    """Gets the information associated with the personal profile,
    such as phone number, city, marital status, and date of birth.

    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :returns: The function returns a dictionary of key/value pairs. If a string \
    is passed in to the info parameter, then the function will return a string \
    corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * user
                      * address
                      * city
                      * state
                      * zipcode
                      * phone_number
                      * marital_status
                      * date_of_birth
                      * citizenship
                      * country_of_residence
                      * number_dependents
                      * signup_as_rhs
                      * tax_id_ssn
                      * updated_at

    """
    url = basic_profile_url()
    data = request_get(url)
    return(filter_data(data, info))


@login_required
def load_investment_profile(info=None):
    """Gets the information associated with the investment profile.
    These are the answers to the questionaire you filled out when you made your profile.

    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :returns: The function returns a dictionary of key/value pairs. \
    If a string is passed in to the info parameter, then the function will return \
    a string corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * user
                      * total_net_worth
                      * annual_income
                      * source_of_funds
                      * investment_objective
                      * investment_experience
                      * liquid_net_worth
                      * risk_tolerance
                      * tax_bracket
                      * time_horizon
                      * liquidity_needs
                      * investment_experience_collected
                      * suitability_verified
                      * option_trading_experience
                      * professional_trader
                      * understand_option_spreads
                      * interested_in_options
                      * updated_at

    """
    url = investment_profile_url()
    data = request_get(url)
    return(filter_data(data, info))


@login_required
def load_portfolio_profile(account_number=None, info=None):
    """Gets the information associated with the portfolios profile,
    such as withdrawable amount, market value of account, and excess margin.

    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :returns: The function returns a dictionary of key/value pairs. \
    If a string is passed in to the info parameter, then the function will return \
    a string corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * url
                      * account
                      * start_date
                      * market_value
                      * equity
                      * extended_hours_market_value
                      * extended_hours_equity
                      * extended_hours_portfolio_equity
                      * last_core_market_value
                      * last_core_equity
                      * last_core_portfolio_equity
                      * excess_margin
                      * excess_maintenance
                      * excess_margin_with_uncleared_deposits
                      * excess_maintenance_with_uncleared_deposits
                      * equity_previous_close
                      * portfolio_equity_previous_close
                      * adjusted_equity_previous_close
                      * adjusted_portfolio_equity_previous_close
                      * withdrawable_amount
                      * unwithdrawable_deposits
                      * unwithdrawable_grants

    """
    url = portfolio_profile_url(account_number)
    if account_number is not None:
        data = request_get(url)
    else:
        data = request_get(url, 'indexzero')
    return(filter_data(data, info))


@login_required
def load_security_profile(info=None):
    """Gets the information associated with the security profile.

    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :returns: The function returns a dictionary of key/value pairs. \
    If a string is passed in to the info parameter, then the function will return \
    a string corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * user
                      * object_to_disclosure
                      * sweep_consent
                      * control_person
                      * control_person_security_symbol
                      * security_affiliated_employee
                      * security_affiliated_firm_relationship
                      * security_affiliated_firm_name
                      * security_affiliated_person_name
                      * security_affiliated_address
                      * security_affiliated_address_subject
                      * security_affiliated_requires_duplicates
                      * stock_loan_consent_status
                      * agreed_to_rhs
                      * agreed_to_rhs_margin
                      * rhs_stock_loan_consent_status
                      * updated_at

    """
    url = security_profile_url()
    data = request_get(url)
    return(filter_data(data, info))


@login_required
def load_user_profile(info=None):
    """Gets the information associated with the user profile,
    such as username, email, and links to the urls for other profiles.

    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :returns: The function returns a dictionary of key/value pairs. \
    If a string is passed in to the info parameter, then the function will return \
    a string corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * url
                      * id
                      * id_info
                      * username
                      * email
                      * email_verified
                      * first_name
                      * last_name
                      * origin
                      * profile_name
                      * created_at

    """
    url = user_profile_url()
    data = request_get(url)
    return(filter_data(data, info))



"""Contains all functions for placing orders for stocks, options, and crypto."""
from uuid import uuid4

from robin_stocks.robinhood.crypto import *
from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.profiles import *
from robin_stocks.robinhood.stocks import *
from robin_stocks.robinhood.urls import *

@login_required
def get_all_stock_orders(info=None, account_number=None, start_date=None):
    """Returns a list of all the orders that have been processed for the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :param start_date: Sets the date of when to start returning orders, returns all orders up to current date and time.
    :type date: Optional[str] format, should this be sent as a DT object? I believe it's safer to require it to be handed to the function as a string.
    :returns: Returns a list of dictionaries of key/value pairs for each order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = orders_url(account_number=account_number, start_date=start_date)
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_all_option_orders(info=None, account_number=None, start_date=None):
    """Returns a list of all the option orders that have been processed for the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :param start_date: Sets the date of when to start returning orders, returns all orders up to current date and time.
    :type date: Optional[str] format, should this be sent as a DT object? I believe it's safer to require it to be handed to the function as a string.
    :returns: Returns a list of dictionaries of key/value pairs for each option order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = option_orders_url(account_number=account_number, start_date=start_date)
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_all_crypto_orders(info=None):
    """Returns a list of all the crypto orders that have been processed for the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each option order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = crypto_orders_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_all_open_stock_orders(info=None, account_number=None):
    """Returns a list of all the orders that are currently open.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = orders_url(account_number=account_number)
    data = request_get(url, 'pagination')

    data = [item for item in data if item['cancel'] is not None]

    return(filter_data(data, info))


@login_required
def get_all_open_option_orders(info=None, account_number=None):
    """Returns a list of all the orders that are currently open.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = option_orders_url(account_number=account_number)
    data = request_get(url, 'pagination')

    data = [item for item in data if item['cancel_url'] is not None]

    return(filter_data(data, info))


@login_required
def get_all_open_crypto_orders(info=None):
    """Returns a list of all the crypto orders that have been processed for the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each option order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = crypto_orders_url()
    data = request_get(url, 'pagination')

    data = [item for item in data if item['cancel_url'] is not None]

    return(filter_data(data, info))


@login_required
def get_stock_order_info(orderID):
    """Returns the information for a single order.

    :param orderID: The ID associated with the order. Can be found using get_all_orders(info=None) or get_all_orders(info=None).
    :type orderID: str
    :returns: Returns a list of dictionaries of key/value pairs for the order.

    """
    url = orders_url(orderID)
    data = request_get(url)
    return(data)


@login_required
def get_option_order_info(order_id):
    """Returns the information for a single option order.

    :param order_id: The ID associated with the option order.
    :type order_id: str
    :returns: Returns a list of dictionaries of key/value pairs for the order.

    """
    url = option_orders_url(order_id)
    data = request_get(url)
    return data


@login_required
def get_crypto_order_info(order_id):
    """Returns the information for a single crypto order.

    :param order_id: The ID associated with the option order.
    :type order_id: str
    :returns: Returns a list of dictionaries of key/value pairs for the order.

    """
    url = crypto_orders_url(order_id)
    data = request_get(url)
    return data


@login_required
def find_stock_orders(**arguments):
    """Returns a list of orders that match the keyword parameters.

    :param arguments: Variable length of keyword arguments. EX. find_orders(symbol='FB',cancel=None,quantity=1)
    :type arguments: str
    :returns: Returns a list of orders.

    """ 
    url = orders_url()
    data = request_get(url, 'pagination')

    if (len(arguments) == 0):
        return(data)

    for item in data:
        item['quantity'] = str(float(item['quantity']))
        item['cumulative_quantity'] = str(float(item['cumulative_quantity']))

    if 'symbol' in arguments.keys():
        arguments['instrument'] = get_instruments_by_symbols(
            arguments['symbol'], info='url')[0]
        del arguments['symbol']

    if 'quantity' in arguments.keys():
        arguments['quantity'] = str(arguments['quantity'])

    stop = len(arguments.keys())-1
    list_of_orders = []
    for item in data:
        for i, (key, value) in enumerate(arguments.items()):
            if key not in item:
                print(error_argument_not_key_in_dictionary(key), file=get_output())
                return([None])
            if value != item[key]:
                break
            if i == stop:
                list_of_orders.append(item)

    return(list_of_orders)


@login_required
def cancel_stock_order(orderID):
    """Cancels a specific order.

    :param orderID: The ID associated with the order. Can be found using get_all_stock_orders(info=None).
    :type orderID: str
    :returns: Returns the order information for the order that was cancelled.

    """ 
    url = cancel_url(orderID)
    data = request_post(url)

    if data:
        print('Order '+str(orderID)+' cancelled', file=get_output())
    return(data)


@login_required
def cancel_option_order(orderID):
    """Cancels a specific option order.

    :param orderID: The ID associated with the order. Can be found using get_all_option_orders(info=None).
    :type orderID: str
    :returns: Returns the order information for the order that was cancelled.

    """ 
    url = option_cancel_url(orderID)
    data = request_post(url)

    if data:
        print('Order '+str(orderID)+' cancelled', file=get_output())
    return(data)


@login_required
def cancel_crypto_order(orderID):
    """Cancels a specific crypto order.

    :param orderID: The ID associated with the order. Can be found using get_all_crypto_orders(info=None).
    :type orderID: str
    :returns: Returns the order information for the order that was cancelled.

    """ 
    url = crypto_cancel_url(orderID)
    data = request_post(url)

    if data:
        print('Order '+str(orderID)+' cancelled', file=get_output())
    return(data)


@login_required
def cancel_all_stock_orders(account_number=None):
    """Cancels all stock orders.

    :returns: The list of orders that were cancelled.

    """ 
    url = orders_url(account_number=account_number)
    data = request_get(url, 'pagination')

    data = [item for item in data if item['cancel'] is not None]

    for item in data:
        request_post(item['cancel'])

    print('All Stock Orders Cancelled', file=get_output())
    return(data)


@login_required
def cancel_all_option_orders(account_number=None):
    """Cancels all option orders.

    :returns: Returns the order information for the orders that were cancelled.

    """ 
    url = option_orders_url(account_number=account_number)
    data = request_get(url, 'pagination')

    data = [item for item in data if item['cancel_url'] is not None]

    for item in data:
        request_post(item['cancel_url'])

    print('All Option Orders Cancelled', file=get_output())
    return(data)


@login_required
def cancel_all_crypto_orders():
    """Cancels all crypto orders.

    :returns: Returns the order information for the orders that were cancelled.

    """ 
    url = crypto_orders_url()
    data = request_get(url, 'pagination')

    data = [item for item in data if item['cancel_url'] is not None]

    for item in data:
        request_post(item['cancel_url'])

    print('All Crypto Orders Cancelled', file=get_output())
    return(data)


@login_required
def order_buy_market(symbol, quantity, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a market order to be executed immediately.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param quantity: The number of stocks to buy.
    :type quantity: int
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "buy", None, None, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_buy_fractional_by_quantity(symbol, quantity, account_number=None, timeInForce='gfd', extendedHours=False, jsonify=True):
    """Submits a market order to be executed immediately for fractional shares by specifying the amount that you want to trade.
    Good for share fractions up to 6 decimal places. Robinhood does not currently support placing limit, stop, or stop loss orders
    for fractional trades.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param quantity: The amount of the fractional shares you want to buy.
    :type quantity: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "buy", None, None, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_buy_fractional_by_price(symbol, amountInDollars, account_number=None, timeInForce='gfd', extendedHours=False, jsonify=True, market_hours='regular_hours'):
    """Submits a market order to be executed immediately for fractional shares by specifying the amount in dollars that you want to trade.
    Good for share fractions up to 6 decimal places. Robinhood does not currently support placing limit, stop, or stop loss orders
    for fractional trades.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param amountInDollars: The amount in dollars of the fractional shares you want to buy.
    :type amountInDollars: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    if amountInDollars < 1:
        print("ERROR: Fractional share price should meet minimum 1.00.", file=get_output())
        return None

    # turn the money amount into decimal number of shares
    price = next(iter(get_latest_price(symbol, 'ask_price', extendedHours)), 0.00)
    fractional_shares = 0 if (price == 0.00) else round_price(amountInDollars/float(price))
    
    return order(symbol, fractional_shares, "buy", None, None, account_number, timeInForce, extendedHours, jsonify, market_hours)


@login_required
def order_buy_limit(symbol, quantity, limitPrice, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a limit order to be executed once a certain price is reached.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param quantity: The number of stocks to buy.
    :type quantity: int
    :param limitPrice: The price to trigger the buy order.
    :type limitPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "buy", limitPrice, None, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_buy_stop_loss(symbol, quantity, stopPrice, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a stop order to be turned into a market order once a certain stop price is reached.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param quantity: The number of stocks to buy.
    :type quantity: int
    :param stopPrice: The price to trigger the market order.
    :type stopPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "buy", None, stopPrice, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_buy_stop_limit(symbol, quantity, limitPrice, stopPrice, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a stop order to be turned into a limit order once a certain stop price is reached.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param quantity: The number of stocks to buy.
    :type quantity: int
    :param limitPrice: The price to trigger the market order.
    :type limitPrice: float
    :param stopPrice: The price to trigger the limit order.
    :type stopPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "buy", limitPrice, stopPrice, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_buy_trailing_stop(symbol, quantity, trailAmount, trailType='percentage', timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a trailing stop buy order to be turned into a market order when traling stop price reached.

    :param symbol: The stock ticker of the stock to buy.
    :type symbol: str
    :param quantity: The number of stocks to buy.
    :type quantity: int
    :param trailAmount: how much to trail by; could be percentage or dollar value depending on trailType
    :type trailAmount: float
    :param trailType: could be "amount" or "percentage"
    :type trailType: str
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.
    """
    return order_trailing_stop(symbol, quantity, "buy", trailAmount, trailType, None, timeInForce, extendedHours, jsonify)


@login_required
def order_sell_market(symbol, quantity, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a market order to be executed immediately.

    :param symbol: The stock ticker of the stock to sell.
    :type symbol: str
    :param quantity: The number of stocks to sell.
    :type quantity: int
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "sell", None, None, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_sell_fractional_by_quantity(symbol, quantity, account_number=None, timeInForce='gfd', priceType='bid_price', extendedHours=False, jsonify=True, market_hours='regular_hours'):
    """Submits a market order to be executed immediately for fractional shares by specifying the amount that you want to trade.
    Good for share fractions up to 6 decimal places. Robinhood does not currently support placing limit, stop, or stop loss orders
    for fractional trades.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param quantity: The amount of the fractional shares you want to buy.
    :type quantity: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "sell", None, None, account_number, timeInForce, extendedHours, jsonify, market_hours)


@login_required
def order_sell_fractional_by_price(symbol, amountInDollars, account_number=None, timeInForce='gfd', extendedHours=False, jsonify=True):
    """Submits a market order to be executed immediately for fractional shares by specifying the amount in dollars that you want to trade.
    Good for share fractions up to 6 decimal places. Robinhood does not currently support placing limit, stop, or stop loss orders
    for fractional trades.

    :param symbol: The stock ticker of the stock to purchase.
    :type symbol: str
    :param amountInDollars: The amount in dollars of the fractional shares you want to buy.
    :type amountInDollars: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    if amountInDollars < 1:
        print("ERROR: Fractional share price should meet minimum 1.00.", file=get_output())
        return None
    # turn the money amount into decimal number of shares
    price = next(iter(get_latest_price(symbol, 'bid_price', extendedHours)), 0.00)
    fractional_shares = 0 if (price == 0.00) else round_price(amountInDollars/float(price))

    return order(symbol, fractional_shares, "sell", None, None, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_sell_limit(symbol, quantity, limitPrice, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a limit order to be executed once a certain price is reached.

    :param symbol: The stock ticker of the stock to sell.
    :type symbol: str
    :param quantity: The number of stocks to sell.
    :type quantity: int
    :param limitPrice: The price to trigger the sell order.
    :type limitPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "sell", limitPrice, None, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_sell_stop_loss(symbol, quantity, stopPrice, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a stop order to be turned into a market order once a certain stop price is reached.

    :param symbol: The stock ticker of the stock to sell.
    :type symbol: str
    :param quantity: The number of stocks to sell.
    :type quantity: int
    :param stopPrice: The price to trigger the market order.
    :type stopPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "sell", None, stopPrice, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_sell_stop_limit(symbol, quantity, limitPrice, stopPrice, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a stop order to be turned into a limit order once a certain stop price is reached.

    :param symbol: The stock ticker of the stock to sell.
    :type symbol: str
    :param quantity: The number of stocks to sell.
    :type quantity: int
    :param limitPrice: The price to trigger the market order.
    :type limitPrice: float
    :param stopPrice: The price to trigger the limit order.
    :type stopPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order(symbol, quantity, "sell", limitPrice, stopPrice, account_number, timeInForce, extendedHours, jsonify)


@login_required
def order_sell_trailing_stop(symbol, quantity, trailAmount, trailType='percentage', timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a trailing stop sell order to be turned into a market order when traling stop price reached.

    :param symbol: The stock ticker of the stock to sell.
    :type symbol: str
    :param quantity: The number of stocks to sell.
    :type quantity: int
    :param trailAmount: how much to trail by; could be percentage or dollar value depending on trailType
    :type trailAmount: float
    :param trailType: could be "amount" or "percentage"
    :type trailType: str
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.
    """
    return order_trailing_stop(symbol, quantity, "sell", trailAmount, trailType, None, timeInForce, extendedHours, jsonify)


@login_required
def order_trailing_stop(symbol, quantity, side, trailAmount, trailType='percentage', account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True):
    """Submits a trailing stop order to be turned into a market order when traling stop price reached.

    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of stocks to trade.
    :type quantity: int
    :param side: buy or sell
    :type side: str
    :param trailAmount: how much to trail by; could be percentage or dollar value depending on trailType
    :type trailAmount: float
    :param trailType: could be "amount" or "percentage"
    :type trailType: str
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: Optional[str]
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """
    try:
        symbol = symbol.upper().strip()
        trailAmount = float(trailAmount)
    except AttributeError as message:
        print(message)
        return None

    stock_price = round_price(get_latest_price(symbol, extendedHours)[0])

    # find stop price based on whether trailType is "amount" or "percentage" and whether its buy or sell
    percentage = 0
    try:
        if trailType == 'amount':
            margin = trailAmount
        else:
            margin = stock_price * trailAmount * 0.01
            percentage = trailAmount
    except Exception as e:
        print('ERROR: {}'.format(e))
        return None

    stopPrice = stock_price + margin if side == "buy" else stock_price - margin
    stopPrice = round_price(stopPrice)

    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'instrument': get_instruments_by_symbols(symbol, info='url')[0],
        'symbol': symbol,
        'quantity': quantity,
        'ref_id': str(uuid4()),
        'type': 'market',
        'stop_price': stopPrice,
        'time_in_force': timeInForce,
        'trigger': 'stop',
        'side': side,
        'extended_hours': extendedHours
    }

    if side == "buy":
        # price should be greater than stopPrice, adding a 5% threshold
        payload['price'] = round_price(stopPrice * 1.05)

    if trailType == 'amount':
        payload['trailing_peg'] = {'type': 'price', 'price': {'amount': trailAmount, 'currency_code': 'USD'}}
    else:
        payload['trailing_peg'] = {'type': 'percentage', 'percentage': str(percentage)}

    url = orders_url(account_number=account_number)
    data = request_post(url, payload, json=True, jsonify_data=jsonify)

    return (data)


@login_required
def order(symbol, quantity, side, limitPrice=None, stopPrice=None, account_number=None, timeInForce='gtc', extendedHours=False, jsonify=True, market_hours='regular_hours'):
    """A generic order function.

    :param symbol: The stock ticker of the stock to sell.
    :type symbol: str
    :param quantity: The number of stocks to sell.
    :type quantity: int
    :param side: Either 'buy' or 'sell'
    :type side: str
    :param limitPrice: The price to trigger the market order.
    :type limitPrice: float
    :param stopPrice: The price to trigger the limit or market order.
    :type stopPrice: float
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day.
    :type timeInForce: str
    :param extendedHours: Premium users only. Allows trading during extended hours. Should be true or false.
    :type extendedHours: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the purchase or selling of stocks, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    orderType = "market"
    trigger = "immediate"

    if side == "buy":
        priceType = "ask_price"
    else:
        priceType = "bid_price"

    if limitPrice and stopPrice:
        price = round_price(limitPrice)
        stopPrice = round_price(stopPrice)
        orderType = "limit"
        trigger = "stop"
    elif limitPrice:
        price = round_price(limitPrice)
        orderType = "limit"
    elif stopPrice:
        stopPrice = round_price(stopPrice)
        if side == "buy":
            price = stopPrice
        else:
            price = None
        trigger = "stop"
    else:
        price = round_price(next(iter(get_latest_price(symbol, priceType, extendedHours)), 0.00))
        
    from datetime import datetime
    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'instrument': get_instruments_by_symbols(symbol, info='url')[0],
        'symbol': symbol,
        'price': price,
        'ask_price': round_price(next(iter(get_latest_price(symbol, "ask_price", extendedHours)), 0.00)),
        'bid_ask_timestamp': datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f'),
        'bid_price': round_price(next(iter(get_latest_price(symbol, "bid_price", extendedHours)), 0.00)),
        'quantity': quantity,
        'ref_id': str(uuid4()),
        'type': orderType,
        'stop_price': stopPrice,
        'time_in_force': timeInForce,
        'trigger': trigger,
        'side': side,
        'market_hours': market_hours, # choices are ['regular_hours', 'all_day_hours', 'extended_hours']
        'extended_hours': extendedHours,
        'order_form_version': 4
    }
    # adjust market orders
    if orderType == 'market':
        if trigger != "stop":
            del payload['stop_price']
        # if market_hours == 'regular_hours': 
        #     del payload['extended_hours'] 
        
    if market_hours == 'regular_hours':
        if side == "buy":
            payload['preset_percent_limit'] = "0.05"
            payload['type'] = 'limit' 
        # regular market sell
        elif orderType == 'market' and side == 'sell':
            del payload['price']   
    elif market_hours in ('extended_hours', 'all_day_hours'):
        payload['type'] = 'limit' 
        payload['quantity']=int(payload['quantity']) # round to integer instead of fractional
        
    url = orders_url(account_number=account_number)
    # print(payload)
    data = request_post(url, payload, jsonify_data=jsonify)

    return(data)


@login_required
def order_option_credit_spread(price, symbol, quantity, spread, timeInForce='gtc', account_number=None, jsonify=True):
    """Submits a limit order for an option credit spread.

    :param price: The limit price to trigger a sell of the option.
    :type price: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to sell.
    :type quantity: int
    :param spread: A dictionary of spread options with the following keys: \n
        - expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.\n
        - strike: The strike price of the option.\n
        - optionType: This should be 'call' or 'put'.\n
        - effect: This should be 'open' or 'close'.\n
        - action: This should be 'buy' or 'sell'.
    :type spread: dict
    :param timeInForce: Changes how long the order will be in effect for. \
     'gtc' = good until cancelled. \
     'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' = execute at opening.
    :type timeInForce: Optional[str]
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the trading of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.
    """
    return(order_option_spread("credit", price, symbol, quantity, spread, timeInForce, account_number, jsonify))


@login_required
def order_option_debit_spread(price, symbol, quantity, spread, timeInForce='gtc', account_number=None, jsonify=True):
    """Submits a limit order for an option debit spread.

    :param price: The limit price to trigger a sell of the option.
    :type price: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to sell.
    :type quantity: int
    :param spread: A dictionary of spread options with the following keys: \n
        - expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.\n
        - strike: The strike price of the option.\n
        - optionType: This should be 'call' or 'put'.\n
        - effect: This should be 'open' or 'close'.\n
        - action: This should be 'buy' or 'sell'.
    :type spread: dict
    :param timeInForce: Changes how long the order will be in effect for.
     'gtc' = good until cancelled. \
     'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' execute at opening.
    :type timeInForce: Optional[str]
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the trading of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.
    """
    return(order_option_spread("debit", price, symbol, quantity, spread, timeInForce, account_number, jsonify))


@login_required
def order_option_spread(direction, price, symbol, quantity, spread, account_number=None, timeInForce='gtc', jsonify=True):
    """Submits a limit order for an option spread. i.e. place a debit / credit spread

    :param direction: Can be "credit" or "debit".
    :type direction: str
    :param price: The limit price to trigger a trade of the option.
    :type price: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to trade.
    :type quantity: int
    :param spread: A dictionary of spread options with the following keys: \n
        - expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.\n
        - strike: The strike price of the option.\n
        - optionType: This should be 'call' or 'put'.\n
        - effect: This should be 'open' or 'close'.\n
        - action: This should be 'buy' or 'sell'.
    :type spread: dict
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for.
     'gtc' = good until cancelled. \
     'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' execute at opening.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the trading of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.
    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None
    legs = []
    for each in spread:
        optionID = id_for_option(symbol,
                                        each['expirationDate'],
                                        each['strike'],
                                        each['optionType'])
        legs.append({'position_effect': each['effect'],
                     'side': each['action'],
                     'ratio_quantity': each['ratio_quantity'],
                     'option': option_instruments_url(optionID)})

    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'direction': direction,
        'time_in_force': timeInForce,
        'legs': legs,
        'type': 'limit',
        'trigger': 'immediate',
        'price': price,
        'quantity': quantity,
        'override_day_trade_checks': False,
        'override_dtbp_checks': False,
        'ref_id': str(uuid4()),
    }

    url = option_orders_url(account_number=account_number)
    data = request_post(url, payload, json=True, jsonify_data=jsonify)

    return(data)


@login_required
def order_buy_option_limit(positionEffect, creditOrDebit, price, symbol, quantity, expirationDate, strike, optionType='both', account_number=None, timeInForce='gtc', jsonify=True):
    """Submits a limit order for an option. i.e. place a long call or a long put.

    :param positionEffect: Either 'open' for a buy to open effect or 'close' for a buy to close effect.
    :type positionEffect: str
    :param creditOrDebit: Either 'debit' or 'credit'.
    :type creditOrDebit: str
    :param price: The limit price to trigger a buy of the option.
    :type price: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to buy.
    :type quantity: int
    :param expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.
    :type expirationDate: str
    :param strike: The strike price of the option.
    :type strike: float
    :param optionType: This should be 'call' or 'put'
    :type optionType: str
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' execute at opening.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    optionID = id_for_option(symbol, expirationDate, strike, optionType)

    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'direction': creditOrDebit,
        'time_in_force': timeInForce,
        'legs': [
            {'position_effect': positionEffect, 'side': 'buy',
                'ratio_quantity': 1, 'option': option_instruments_url(optionID)},
        ],
        'type': 'limit',
        'trigger': 'immediate',
        'price': price,
        'quantity': quantity,
        'override_day_trade_checks': False,
        'override_dtbp_checks': False,
        'ref_id': str(uuid4()),
    }

    url = option_orders_url(account_number=account_number)
    # print(payload)
    data = request_post(url, payload, json=True, jsonify_data=jsonify)

    return(data)


@login_required
def order_buy_option_stop_limit(positionEffect, creditOrDebit, limitPrice, stopPrice, symbol, quantity, expirationDate, strike, optionType='both', account_number=None, timeInForce='gtc', jsonify=True):
    """Submits a stop order to be turned into a limit order once a certain stop price is reached.

    :param positionEffect: Either 'open' for a buy to open effect or 'close' for a buy to close effect.
    :type positionEffect: str
    :param creditOrDebit: Either 'debit' or 'credit'.
    :type creditOrDebit: str
    :param limitPrice: The limit price to trigger a buy of the option.
    :type limitPrice: float
    :param stopPrice: The price to trigger the limit order.
    :type stopPrice: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to buy.
    :type quantity: int
    :param expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.
    :type expirationDate: str
    :param strike: The strike price of the option.
    :type strike: float
    :param optionType: This should be 'call' or 'put'
    :type optionType: str
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' execute at opening.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    optionID = id_for_option(symbol, expirationDate, strike, optionType)

    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'direction': creditOrDebit,
        'time_in_force': timeInForce,
        'legs': [
            {'position_effect': positionEffect, 'side': 'buy',
                'ratio_quantity': 1, 'option': option_instruments_url(optionID)},
        ],
        'type': 'limit',
        'trigger': 'stop',
        'price': limitPrice,
        'stop_price': stopPrice,
        'quantity': quantity,
        'override_day_trade_checks': False,
        'override_dtbp_checks': False,
        'ref_id': str(uuid4()),
    }

    url = option_orders_url(account_number=account_number)
    data = request_post(url, payload, json=True, jsonify_data=jsonify)

    return(data)


def order_sell_option_stop_limit(positionEffect, creditOrDebit, limitPrice, stopPrice, symbol, quantity, expirationDate, strike, optionType='both', account_number=None, timeInForce='gtc', jsonify=True):
    """Submits a stop order to be turned into a limit order once a certain stop price is reached.

    :param positionEffect: Either 'open' for a buy to open effect or 'close' for a buy to close effect.
    :type positionEffect: str
    :param creditOrDebit: Either 'debit' or 'credit'.
    :type creditOrDebit: str
    :param limitPrice: The limit price to trigger a buy of the option.
    :type limitPrice: float
    :param stopPrice: The price to trigger the limit order.
    :type stopPrice: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to buy.
    :type quantity: int
    :param expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.
    :type expirationDate: str
    :param strike: The strike price of the option.
    :type strike: float
    :param optionType: This should be 'call' or 'put'
    :type optionType: str
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' execute at opening.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    optionID = id_for_option(symbol, expirationDate, strike, optionType)

    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'direction': creditOrDebit,
        'time_in_force': timeInForce,
        'legs': [
            {'position_effect': positionEffect, 'side': 'sell',
                'ratio_quantity': 1, 'option': option_instruments_url(optionID)},
        ],
        'type': 'limit',
        'trigger': 'stop',
        'price': limitPrice,
        'stop_price': stopPrice,
        'quantity': quantity,
        'override_day_trade_checks': False,
        'override_dtbp_checks': False,
        'ref_id': str(uuid4()),
    }

    url = option_orders_url(account_number=account_number)
    data = request_post(url, payload, json=True, jsonify_data=jsonify)

    return(data)


@login_required
def order_sell_option_limit(positionEffect, creditOrDebit, price, symbol, quantity, expirationDate, strike, optionType='both', account_number=None, timeInForce='gtc', jsonify=True):
    """Submits a limit order for an option. i.e. place a short call or a short put.

    :param positionEffect: Either 'open' for a sell to open effect or 'close' for a sell to close effect.
    :type positionEffect: str
    :param creditOrDebit: Either 'debit' or 'credit'.
    :type creditOrDebit: str
    :param price: The limit price to trigger a sell of the option.
    :type price: float
    :param symbol: The stock ticker of the stock to trade.
    :type symbol: str
    :param quantity: The number of options to sell.
    :type quantity: int
    :param expirationDate: The expiration date of the option in 'YYYY-MM-DD' format.
    :type expirationDate: str
    :param strike: The strike price of the option.
    :type strike: float
    :param optionType: This should be 'call' or 'put'
    :type optionType: str
    :param account_number: the robinhood account number.
    :type account_number: Optional[str]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled. \
    'gfd' = good for the day. 'ioc' = immediate or cancel. 'opg' execute at opening.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of options, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    optionID = id_for_option(symbol, expirationDate, strike, optionType)

    payload = {
        'account': load_account_profile(account_number=account_number, info='url'),
        'direction': creditOrDebit,
        'time_in_force': timeInForce,
        'legs': [
            {'position_effect': positionEffect, 'side': 'sell',
                'ratio_quantity': 1, 'option': option_instruments_url(optionID)},
        ],
        'type': 'limit',
        'trigger': 'immediate',
        'price': price,
        'quantity': quantity,
        'override_day_trade_checks': False,
        'override_dtbp_checks': False,
        'ref_id': str(uuid4()),
    }

    url = option_orders_url(account_number=account_number)
    data = request_post(url, payload, json=True, jsonify_data=jsonify)

    return(data)


@login_required
def order_buy_crypto_by_price(symbol, amountInDollars, timeInForce='gtc', jsonify=True):
    """Submits a market order for a crypto by specifying the amount in dollars that you want to trade.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param amountInDollars: The amount in dollars of the crypto you want to buy.
    :type amountInDollars: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order_crypto(symbol, "buy", amountInDollars, "price", None, timeInForce, jsonify)


@login_required
def order_buy_crypto_by_quantity(symbol, quantity, timeInForce='gtc', jsonify=True):
    """Submits a market order for a crypto by specifying the decimal amount of shares to buy.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param quantity: The decimal amount of shares to buy.
    :type quantity: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order_crypto(symbol, "buy", quantity, "quantity", None, timeInForce, jsonify)


@login_required
def order_buy_crypto_limit(symbol, quantity, limitPrice, timeInForce='gtc', jsonify=True):
    """Submits a limit order for a crypto by specifying the decimal amount of shares to buy.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param quantity: The decimal amount of shares to buy.
    :type quantity: float
    :param limitPrice: The limit price to set for the crypto.
    :type limitPrice: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order_crypto(symbol, "buy", quantity, "quantity", limitPrice, timeInForce, jsonify)


@login_required
def order_buy_crypto_limit_by_price(symbol, amountInDollars, limitPrice, timeInForce='gtc', jsonify=True):
    """Submits a limit order for a crypto by specifying the decimal price to buy.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param amountInDollars: The amount in dollars of the crypto you want to buy.
    :type amountInDollars: float
    :param limitPrice: The limit price to set for the crypto.
    :type limitPrice: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """
    return order_crypto(symbol, "buy", amountInDollars, "price", limitPrice, timeInForce, jsonify)


@login_required
def order_sell_crypto_by_price(symbol, amountInDollars, timeInForce='gtc', jsonify=True):
    """Submits a market order for a crypto by specifying the amount in dollars that you want to trade.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param amountInDollars: The amount in dollars of the crypto you want to sell.
    :type amountInDollars: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order_crypto(symbol, "sell", amountInDollars, "price", None, timeInForce, jsonify)


@login_required
def order_sell_crypto_by_quantity(symbol, quantity, timeInForce='gtc', jsonify=True):
    """Submits a market order for a crypto by specifying the decimal amount of shares to buy.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param quantity: The decimal amount of shares to sell.
    :type quantity: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """ 
    return order_crypto(symbol, "sell", quantity, "quantity", None, timeInForce, jsonify)


@login_required
def order_sell_crypto_limit(symbol, quantity, limitPrice, timeInForce='gtc', jsonify=True):
    """Submits a limit order for a crypto by specifying the decimal amount of shares to sell.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param quantity: The decimal amount of shares to sell.
    :type quantity: float
    :param limitPrice: The limit price to set for the crypto.
    :type limitPrice: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """
    return order_crypto(symbol, "sell", quantity, "quantity", limitPrice, timeInForce, jsonify)


@login_required
def order_sell_crypto_limit_by_price(symbol, amountInDollars, limitPrice, timeInForce='gtc', jsonify=True):
    """Submits a limit order for a crypto by specifying the decimal price to sell.
    Good for share fractions up to 8 decimal places.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param amountInDollars: The amount in dollars of the crypto you want to sell.
    :type amountInDollars: float
    :param limitPrice: The limit price to set for the crypto.
    :type limitPrice: float
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the buying of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """
    return order_crypto(symbol, "sell", amountInDollars, "price", limitPrice, timeInForce, jsonify)


@login_required
def order_crypto(symbol, side, quantityOrPrice, amountIn="quantity", limitPrice=None, timeInForce="gtc", jsonify=True):
    """Submits an order for a crypto.

    :param symbol: The crypto ticker of the crypto to trade.
    :type symbol: str
    :param side: Either 'buy' or 'sell'
    :type side: str
    :param quantityOrPrice: Either the decimal price of shares to trade or the decimal quantity of shares.
    :type quantityOrPrice: float
    :param amountIn: If left default value of 'quantity', order will attempt to trade cryptos by the amount of crypto \
        you want to trade. If changed to 'price', order will attempt to trade cryptos by the price you want to buy or sell.
    :type amountIn: Optional[str]
    :param limitPrice: The price to trigger the market order.
    :type limitPrice: Optional[float]
    :param timeInForce: Changes how long the order will be in effect for. 'gtc' = good until cancelled.
    :type timeInForce: Optional[str]
    :param jsonify: If set to False, function will return the request object which contains status code and headers.
    :type jsonify: Optional[str]
    :returns: Dictionary that contains information regarding the selling of crypto, \
    such as the order id, the state of order (queued, confired, filled, failed, canceled, etc.), \
    the price, and the quantity.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    crypto_id = get_crypto_id(symbol)
    orderType = "market"

    if side == "buy":
        priceType = "ask_price"
    else:
        priceType = "bid_price"

    if limitPrice:
        price = limitPrice
        orderType = "limit"
    else:
        price = round_price(get_crypto_quote_from_id(crypto_id, info=priceType))

    if amountIn == "quantity":
        quantity = quantityOrPrice
    else:
        quantity = round_price(quantityOrPrice/price)

    payload = {
        'account_id': load_crypto_profile(info="id"),
        'currency_pair_id': crypto_id,
        'price': price,
        'quantity': quantity,
        'ref_id': str(uuid4()),
        'side': side,
        'time_in_force': timeInForce,
        'type': orderType
    }

    url = order_crypto_url()

    # This is safe because 'ref_id' guards us from duplicate orders
    attempts = 3
    while attempts > 0:
        data = request_post(url, payload, json=True, jsonify_data=jsonify)
        if data is not None:
            break

        attempts -= 1

    return(data)



from .account import (check_available_balances, check_notional_balances,
                      check_transfers, get_account_detail,
                      get_approved_addresses, get_deposit_addresses,
                      withdraw_crypto_funds)
from .authentication import heartbeat, login, logout
from .crypto import (get_notional_volume, get_price, get_pubticker,
                     get_symbol_details, get_symbols, get_ticker,
                     get_trade_volume)
from .helper import (get_login_state, request_get, set_default_json_flag,
                     use_sand_box_urls)
from .orders import (active_orders, cancel_all_active_orders,
                     cancel_all_session_orders, cancel_order,
                     get_trades_for_crypto, order, order_market, order_status)



import pickle
from datetime import datetime, timedelta
from pathlib import Path

from cryptography.fernet import Fernet
from robin_stocks.tda.globals import DATA_DIR_NAME, PICKLE_NAME
from robin_stocks.tda.helper import (request_data, set_login_state,
                                     update_session)
from robin_stocks.tda.urls import URLS


def login_first_time(encryption_passcode, client_id, authorization_token, refresh_token):
    """ Stores log in information in a pickle file on the computer. After being used once,
    user can call login() to automatically read in information from pickle file and refresh
    authorization tokens when needed.

    :param encryption_passcode: Encryption key created by generate_encryption_passcode().
    :type encryption_passcode: str
    :param client_id: The Consumer Key for the API account.
    :type client_id: str
    :param authorization_token: The authorization code returned from post request to https://developer.tdameritrade.com/authentication/apis/post/token-0
    :type authorization_token: str
    :param refresh_token: The refresh code returned from post request to https://developer.tdameritrade.com/authentication/apis/post/token-0
    :type refresh_token: str

    """
    if type(encryption_passcode) is str:
        encryption_passcode = encryption_passcode.encode()
    cipher_suite = Fernet(encryption_passcode)
    # Create necessary folders and paths for pickle file as defined in globals.
    data_dir = Path.home().joinpath(DATA_DIR_NAME)
    if not data_dir.exists():
        data_dir.mkdir(parents=True)
    pickle_path = data_dir.joinpath(PICKLE_NAME)
    if not pickle_path.exists():
        Path.touch(pickle_path)
    # Write information to the file.
    with pickle_path.open("wb") as pickle_file:
        pickle.dump(
            {
                'authorization_token': cipher_suite.encrypt(authorization_token.encode()),
                'refresh_token': cipher_suite.encrypt(refresh_token.encode()),
                'client_id': cipher_suite.encrypt(client_id.encode()),
                'authorization_timestamp': datetime.now(),
                'refresh_timestamp': datetime.now()
            }, pickle_file)


def login(encryption_passcode):
    """ Set the authorization token so the API can be used. Gets a new authorization token
    every 30 minutes using the refresh token. Gets a new refresh token every 60 days.

    :param encryption_passcode: Encryption key created by generate_encryption_passcode().
    :type encryption_passcode: str
    
    """
    if type(encryption_passcode) is str:
        encryption_passcode = encryption_passcode.encode()
    cipher_suite = Fernet(encryption_passcode)
    # Check that file exists before trying to read from it.
    data_dir = Path.home().joinpath(DATA_DIR_NAME)
    pickle_path = data_dir.joinpath(PICKLE_NAME)
    if not pickle_path.exists():
        raise FileExistsError(
            "Please Call login_first_time() to create pickle file.")
    # Read the information from the pickle file.
    with pickle_path.open("rb") as pickle_file:
        pickle_data = pickle.load(pickle_file)
        access_token = cipher_suite.decrypt(pickle_data['authorization_token']).decode()
        refresh_token = cipher_suite.decrypt(pickle_data['refresh_token']).decode()
        client_id = cipher_suite.decrypt(pickle_data['client_id']).decode()
        authorization_timestamp = pickle_data['authorization_timestamp']
        refresh_timestamp = pickle_data['refresh_timestamp']
    # Authorization tokens expire after 30 mins. Refresh tokens expire after 90 days,
    # but you need to request a fresh authorization and refresh token before it expires.
    authorization_delta = timedelta(seconds=1800)
    refresh_delta = timedelta(days=60)
    url = URLS.oauth()
    # If it has been longer than 60 days. Get a new refresh and authorization token.
    # Else if it has been longer than 30 minutes, get only a new authorization token.
    if (datetime.now() - refresh_timestamp > refresh_delta):
        payload = {
            "grant_type": "refresh_token",
            "access_type": "offline",
            "refresh_token": refresh_token,
            "client_id": client_id
        }
        data, _ = request_data(url, payload, True)
        if "access_token" not in data and "refresh_token" not in data:
            raise ValueError(
                "Refresh token is no longer valid. Call login_first_time() to get a new refresh token.")
        access_token = data["access_token"]
        refresh_token = data["refresh_token"]
        with pickle_path.open("wb") as pickle_file:
            pickle.dump(
                {
                    'authorization_token': cipher_suite.encrypt(access_token.encode()),
                    'refresh_token': cipher_suite.encrypt(refresh_token.encode()),
                    'client_id': cipher_suite.encrypt(client_id.encode()),
                    'authorization_timestamp': datetime.now(),
                    'refresh_timestamp': datetime.now()
                }, pickle_file)
    elif (datetime.now() - authorization_timestamp > authorization_delta):
        payload = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": client_id
        }
        data, _ = request_data(url, payload, True)
        if "access_token" not in data:
            raise ValueError(
                "Refresh token is no longer valid. Call login_first_time() to get a new refresh token.")
        access_token = data["access_token"]
        # Write new data to file. Do not replace the refresh timestamp.
        with pickle_path.open("wb") as pickle_file:
            pickle.dump(
                {
                    'authorization_token': cipher_suite.encrypt(access_token.encode()),
                    'refresh_token': cipher_suite.encrypt(refresh_token.encode()),
                    'client_id': cipher_suite.encrypt(client_id.encode()),
                    'authorization_timestamp': datetime.now(),
                    'refresh_timestamp': refresh_timestamp
                }, pickle_file)
    # Store authorization token in session information to be used with API calls.
    auth_token = "Bearer {0}".format(access_token)
    update_session("Authorization", auth_token)
    update_session("apikey", client_id)
    set_login_state(True)
    return auth_token


def generate_encryption_passcode():
    """ Returns an encryption key to be used for logging in.

    :returns: Returns a byte object to be used with cryptography.

    """
    return Fernet.generate_key().decode()
